from typing import List
import torch

# accessed in global scope to track non-parameter packed bytes (npp) as a simple proxy (ceiling) for activation memory
_npp_bytes = 0

def _hook_npp_pre_forward(module, *args, **kwargs):
    mem = module.mem_info_handle()
    global _npp_bytes
    module.npp_pre_forward = _npp_bytes
    module.rss_pre_forward = mem.rss
    return None

def _hook_npp_post_forward(module, *args, **kwargs):
    global _npp_bytes
    module.npp_post_forward = _npp_bytes
    module.npp_diff = module.npp_post_forward - module.npp_pre_forward
    mem = module.mem_info_handle()
    module.rss_post_forward = mem.rss
    rss_diff = module.rss_post_forward - module.rss_pre_forward
    module.rss_diff = rss_diff + (module.rss_diff if hasattr(module, "rss_diff") else 0)
    return None

def _reset_memory_hooks_state(model, reset_attrs: List[str]):
    global _npp_bytes
    _npp_bytes = 0
    for module in model.modules():
        for attr in reset_attrs:
            setattr(module, attr, 0)

def _npp_hook(x):
    global _npp_bytes
    if not isinstance(x, torch.nn.Parameter):
        _npp_bytes += x.nbytes
    return x



from typing import Optional
import torch

from fts_examples.profiling import MemProfiler, MemProfilerCfg


class ProfilerHooksMixin:
    """Initially supporting only MemProfiler but will likely be extended to support more profiling in the
    future."""
    def __init__(self, memprofiler_cfg: Optional[MemProfilerCfg] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.memprofiler = None
        self.memprofiler_cfg = memprofiler_cfg
        if self.memprofiler_cfg and self.memprofiler_cfg.enabled:
            self.memprofiler = MemProfiler()
            self.memprofiler.connect(self)
            if self.cuda_allocator_history:
                torch.cuda.memory._record_memory_history()

    def setup(self, stage):
        super().setup(stage)
        if self.memprofiler:
            self.memprofiler.init_memprof_log_dir()

    def on_train_start(self) -> None:
        if self.memprofiler:
            self.memprofiler.maybe_init_fsdp_mem_tracker()
        super().on_train_start()

    def on_train_end(self) -> None:
        if self.memprofiler:
            self.memprofiler.dump_memory_stats()
        super().on_train_end()

    @property
    def cuda_allocator_history(self) -> bool:
        return self.memprofiler and self.memprofiler_cfg.cuda_allocator_history

    @property
    def memprofiler(self) -> Optional[MemProfiler]:
        """Returns the MemProfiler instance if it has been initialized, otherwise None."""
        return getattr(self, '_memprofiler', None)

    @memprofiler.setter
    def memprofiler(self, memprofiler: Optional[MemProfiler]) -> None:
        """Sets the MemProfiler instance if it is not already initialized."""
        if not getattr(self, '_memprofiler', None):
            setattr(self, '_memprofiler', memprofiler)



from fts_examples.patching._patch_utils import _prepare_module_ctx
from lightning.pytorch.cli import LightningCLI  # noqa: F401

# we ignore these for the entire file since we're using our global namespace trickeration to patch
# ruff: noqa: F821
# pyright: reportUndefinedVariable=false

globals().update(_prepare_module_ctx('lightning.pytorch.cli', globals()))

def _updated_parse_known_args_patch(self: ArgumentParser, args: Any = None, namespace: Any = None,
                                    intermixed: bool = False) -> tuple[Any, Any]:
    namespace, args = super(ArgumentParser, self)._parse_known_args(args, namespace,
                                                                    intermixed=intermixed)  # type: ignore
    return namespace, args



import operator
import sys
import os
from enum import Enum
from typing import NamedTuple, Tuple, Callable
from fts_examples.patching._patch_utils import lwt_compare_version
from lightning.pytorch.cli import _JSONARGPARSE_SIGNATURES_AVAILABLE

class OSEnvToggle(NamedTuple):
    env_var_name: str
    default: str

class DependencyPatch(NamedTuple):
    """Ephemeral dependency patches to conditionally apply to the environment.

    To activate a given patch, all defined `condition` callables must return truthy and the `env_flag` must be set (or
    must default) to '1'
    """
    condition: Tuple[Callable]  # typically a tuple of `lwt_compare_version` to define version dependency
    env_flag: OSEnvToggle  # a tuple defining the environment variable based condition and its default if not set
    function: Callable
    patched_package: str
    description: str


def _dep_patch_repr(self):
    return f'Patch of {self.patched_package}: {self.description})'

DependencyPatch.__repr__ = _dep_patch_repr

# N.B. One needs to ensure they patch all relevant _calling module_ references to patch targets since we usually patch
# after those calling modules have already secured the original (unpatched) references.

def _patch_einsum_strategies():
    from fts_examples.patching.patched_einsum_strategies import gen_einsum_strategies

    # In this case fortunately, we only import/call `gen_einsum_strategies` from
    # `torch.distributed._tensor.ops.matrix_ops`, so only need to patch there.
    target_mod = 'torch.distributed.tensor._ops._matrix_ops'
    sys.modules.get(target_mod).__dict__['gen_einsum_strategies'] = gen_einsum_strategies

def _patch_unsupported_numpy_arrow_extractor():
    from fts_examples.patching.patched_numpyarrowextractor import NumpyArrowExtractor
    # since the TorchFormatter and NumpyFormatter classes are already defined we need to patch both definitions
    # to use our patched `NumpyArrowExtractor`
    for old_mod, stale_ref in zip(['torch_formatter', 'np_formatter'], ['TorchFormatter', 'NumpyFormatter']):
        target_mod = f'datasets.formatting.{old_mod}'
        sys.modules.get(target_mod).__dict__.get(stale_ref).numpy_arrow_extractor = NumpyArrowExtractor


def _patch_triton():
    from fts_examples.patching.patched_triton_jit_fn_init import _new_init
    target_mod = 'triton.runtime.jit'
    sys.modules.get(target_mod).__dict__.get('JITFunction').__init__ = _new_init

def _patch_lightning_jsonargparse():
    from fts_examples.patching.patched_lightning_jsonargparse import _updated_parse_known_args_patch
    target_mod = 'lightning.pytorch.cli'
    sys.modules.get(target_mod).__dict__.get('ArgumentParser')._parse_known_args = _updated_parse_known_args_patch

# TODO: remove once `2.6.0` is minimum
# required for `torch==2.5.x`, TBD wrt subsequent versions though appears fixed in torch `2.6.0` nightlies
einsum_strategies_patch = DependencyPatch(
    condition=(lwt_compare_version("torch", operator.le, "2.5.2"),
               lwt_compare_version("torch", operator.ge, "2.5.0"),),
    env_flag=OSEnvToggle("ENABLE_FTS_EINSUM_STRATEGY_PATCH", default="0"),
    function=_patch_einsum_strategies, patched_package='torch',
    description='Address trivial tp submesh limitation until PyTorch provides upstream fix')

# TODO: remove when min jsonargparse is `4.35.0` or another `2.5.0.postN` lightning release fixes
lightning_jsonargparse_patch = DependencyPatch(
    condition=(sys.version_info >= (3, 12, 8), lwt_compare_version("jsonargparse", operator.lt, "4.35.0"),
               _JSONARGPARSE_SIGNATURES_AVAILABLE),
               env_flag=OSEnvToggle("ENABLE_FTS_LIGHTNING_JSONARGPARSE_PATCH", default="1"),
               function=_patch_lightning_jsonargparse,
               patched_package='lightning',
               description=("For the edge case where `lightning` allows a `jsonargparse` version that doesn't support"
                            " python versions >= `3.12.8`.")
)

# TODO: remove once `datasets==2.21.0` is minimum
datasets_numpy_extractor_patch = DependencyPatch(
    condition=(lwt_compare_version("numpy", operator.ge, "2.0.0"),
               lwt_compare_version("datasets", operator.lt, "2.21.0")),
               env_flag=OSEnvToggle("ENABLE_FTS_NUMPY_EXTRACTOR_PATCH", default="1"),
               function=_patch_unsupported_numpy_arrow_extractor,
               patched_package='datasets',
               description='Adjust `NumpyArrowExtractor` to properly use `numpy` 2.0 copy semantics')

# TODO: remove once `torch 2.5.0` is minimum, only required for `torch==2.4.x`
triton_codgen_patch = DependencyPatch(
    condition=(lwt_compare_version("pytorch-triton", operator.eq, "3.0.0", "45fff310c8"),),
    env_flag=OSEnvToggle("ENABLE_FTS_TRITON_CODEGEN_PATCH", default="1"),
    function=_patch_triton, patched_package='pytorch-triton',
    description='Address `triton` #3564 until PyTorch pins the upstream fix')

class ExpPatch(Enum):
    EINSUM_STRATEGIES = einsum_strategies_patch
    NUMPY_EXTRACTOR = datasets_numpy_extractor_patch
    TRITON_CODEGEN = triton_codgen_patch
    LIGHTNING_JSONARGPARSE = lightning_jsonargparse_patch

_DEFINED_PATCHES = set(ExpPatch)
_ACTIVE_PATCHES = set()

for defined_patch in _DEFINED_PATCHES:
    if all(defined_patch.value.condition) and os.environ.get(defined_patch.value.env_flag.env_var_name,
                                                             defined_patch.value.env_flag.default) == "1":
        defined_patch.value.function()
        _ACTIVE_PATCHES.add(defined_patch)



from fts_examples.patching._patch_utils import _prepare_module_ctx
from lightning.fabric.utilities.imports import _NUMPY_AVAILABLE

# we ignore these for the entire file since we're using our global namespace trickeration to patch
# ruff: noqa: F821
# pyright: reportUndefinedVariable=false

if _NUMPY_AVAILABLE:
    globals().update(_prepare_module_ctx('datasets.formatting.formatting', globals()))


    class NumpyArrowExtractor(BaseArrowExtractor[dict, np.ndarray, dict]):

        def __init__(self, **np_array_kwargs):
            self.np_array_kwargs = np_array_kwargs

        def extract_row(self, pa_table: pa.Table) -> dict:
            return _unnest(self.extract_batch(pa_table))

        def extract_column(self, pa_table: pa.Table) -> np.ndarray:
            return self._arrow_array_to_numpy(pa_table[pa_table.column_names[0]])

        def extract_batch(self, pa_table: pa.Table) -> dict:
            return {col: self._arrow_array_to_numpy(pa_table[col]) for col in pa_table.column_names}

        def _arrow_array_to_numpy(self, pa_array: pa.Array) -> np.ndarray:
            if isinstance(pa_array, pa.ChunkedArray):
                if isinstance(pa_array.type, _ArrayXDExtensionType):
                    # don't call to_pylist() to preserve dtype of the fixed-size array
                    zero_copy_only = _is_zero_copy_only(pa_array.type.storage_dtype, unnest=True)
                    array: List = [
                        row for chunk in pa_array.chunks for row in chunk.to_numpy(zero_copy_only=zero_copy_only)
                    ]
                else:
                    zero_copy_only = _is_zero_copy_only(pa_array.type) and all(
                        not _is_array_with_nulls(chunk) for chunk in pa_array.chunks
                    )
                    array: List = [
                        row for chunk in pa_array.chunks for row in chunk.to_numpy(zero_copy_only=zero_copy_only)
                    ]
            else:
                if isinstance(pa_array.type, _ArrayXDExtensionType):
                    # don't call to_pylist() to preserve dtype of the fixed-size array
                    zero_copy_only = _is_zero_copy_only(pa_array.type.storage_dtype, unnest=True)
                    array: List = pa_array.to_numpy(zero_copy_only=zero_copy_only)
                else:
                    zero_copy_only = _is_zero_copy_only(pa_array.type) and not _is_array_with_nulls(pa_array)
                    array: List = pa_array.to_numpy(zero_copy_only=zero_copy_only).tolist()
            if len(array) > 0:
                if any(
                    (isinstance(x, np.ndarray) and (x.dtype == object or x.shape != array[0].shape))
                    or (isinstance(x, float) and np.isnan(x))
                    for x in array
                ):
                    return np.array(array, copy=False, dtype=object)
            # return np.array(array, copy=False)
            return np.asarray(array)



from fts_examples.patching._patch_utils import _prepare_module_ctx
import re
from triton.runtime import jit  # noqa: F401


globals().update(_prepare_module_ctx('triton.runtime.jit', globals()))

# we ignore these for the entire file since we're using our global namespace trickeration to patch
# ruff: noqa: F821
# pyright: reportUndefinedVariable=false


def _new_init(self, fn, version=None, do_not_specialize=None, debug=None, noinline=None, repr=None,
              launch_metadata=None):
    do_not_specialize = do_not_specialize if do_not_specialize else []

    self.fn = fn
    self.module = fn.__module__
    self.version = version
    self.signature = inspect.signature(fn)
    self.do_not_specialize = do_not_specialize
    self.starting_line_number = inspect.getsourcelines(fn)[1]
    self.repr = lambda _: fn.__name__ if repr is None else repr(_)
    self.launch_metadata = launch_metadata

    self.params = []
    for i, param in enumerate(self.signature.parameters.values()):
        dns = do_not_specialize and (i in do_not_specialize or param.name in do_not_specialize)
        self.params.append(KernelParam(i, param, dns))

    # function source code (without decorators)
    self.src = textwrap.dedent(inspect.getsource(fn))
    #self.src = self.src[self.src.find("def"):]
    self.src = self.src[re.search(r"^def\s+\w+\s*\(", self.src, re.MULTILINE).start():]
    # cache of just-in-time compiled kernels
    self.cache = defaultdict(dict)
    self.hash = None
    # JITFunction can be instantiated as kernel
    # when called with a grid using __getitem__
    self.kernel = None
    self.debug = True if os.environ.get("TRITON_DEBUG", "0") == "1" else debug
    self.noinline = noinline

    # TODO(jlebar): Remove uses of these fields outside this file, then
    # remove the fields here.
    self.arg_names = [p.name for p in self.params]
    self.constexprs = [p.num for p in self.params if p.is_constexpr]

    # Hooks that will be called prior to executing "run"
    self.pre_run_hooks = []

    # reuse docs of wrapped function
    self.__doc__ = fn.__doc__
    self.__name__ = fn.__name__
    self.__globals__ = fn.__globals__
    self.__module__ = fn.__module__

# JITFunction.__init__ = _new_init



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#
# Llama 2 is licensed under the LLAMA 2 Community License,
# Copyright (c) Meta Platforms, Inc. All Rights Reserved.

# This is a slightly modified version of https://bit.ly/torchtitan_llama_d2a4904
# The only changes are:
# - we fix use of `RMSNorm` rather than parameterizing the `norm_type` to avoid the need to import another
#   torchtitan module (`torchtitan.models.norms`: https://bit.ly/torchtitan_norms_module)
# - we add a `reset_parameters` method that invokes `init_weights` to Transformers for more convenient deferred
#   materialization (required for `freqs_cis` buffer)

from dataclasses import dataclass
from typing import Optional, Tuple

import torch
import torch.nn.functional as F
from torch import nn


@dataclass
class ModelCfg:
    dim: int = 4096
    n_layers: int = 32
    n_heads: int = 32
    n_kv_heads: Optional[int] = None
    vocab_size: int = -1  # defined later by tokenizer
    multiple_of: int = 256  # make SwiGLU hidden layer size multiple of large power of 2
    ffn_dim_multiplier: Optional[float] = None
    norm_eps: float = 1e-5
    rope_theta: float = 10000
    max_batch_size: int = 32
    max_seq_len: int = 2048
    # If `True`, then each transformer block init uses its layer ID, and if
    # `False`, each uses the total number of transformer blocks
    depth_init: bool = True


class RMSNorm(nn.Module):
    """Initialize the RMSNorm normalization layer.

    Args:
        dim (int): The dimension of the input tensor.
        eps (float, optional): A small value added to the denominator for numerical stability. Default is 1e-6.

    Attributes:
        eps (float): A small value added to the denominator for numerical stability.
        weight (nn.Parameter): Learnable scaling parameter.
    """

    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x: torch.Tensor):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x: torch.Tensor):
        output = self._norm(x.float()).type_as(x)
        return output * self.weight

    def reset_parameters(self):
        torch.nn.init.ones_(self.weight)  # type: ignore


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0) -> torch.Tensor:
    """Precompute the frequency tensor for complex exponentials (cis) with given dimensions.

    This function calculates a frequency tensor with complex exponentials using the given dimension 'dim'
    and the end index 'end'. The 'theta' parameter scales the frequencies.
    The returned tensor contains complex values in complex64 data type.

    Args:
        dim (int): Dimension of the frequency tensor.
        end (int): End index for precomputing frequencies.
        theta (float, optional): Scaling factor for frequency computation. Defaults to 10000.0.

    Returns:
        torch.Tensor: Precomputed frequency tensor with complex exponentials.
    """
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
    t = torch.arange(end, device=freqs.device)
    freqs = torch.outer(t, freqs).float()
    freqs_cis = torch.polar(torch.ones_like(freqs), freqs)  # complex64
    return freqs_cis


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
    """Reshape frequency tensor for broadcasting it with another tensor.

    This function reshapes the frequency tensor to have the same shape as the target tensor 'x'
    for the purpose of broadcasting the frequency tensor during element-wise operations.

    The input freqs_cis tensor is assumed to be of shape (max_seqlen, dim),
    and the first seqlen elements will be sliced, but dim must match x.

    Args:
        freqs_cis (torch.Tensor): Frequency tensor to be reshaped.
        x (torch.Tensor): Target tensor for broadcasting compatibility.

    Returns:
        torch.Tensor: Reshaped frequency tensor.
    """
    ndim = x.ndim
    assert 0 <= 1 < ndim
    seqlen = x.shape[1]
    freqs_cis = freqs_cis[0:seqlen]
    assert freqs_cis.shape == (seqlen, x.shape[-1])
    shape = [d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)]
    return freqs_cis.view(*shape)


def apply_rotary_emb(
    xq: torch.Tensor,
    xk: torch.Tensor,
    freqs_cis: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Apply rotary embeddings to input tensors using the given frequency tensor.

    This function applies rotary embeddings to the given query 'xq' and key 'xk' tensors using the provided
    frequency tensor 'freqs_cis'. The input tensors are reshaped as complex numbers, and the frequency tensor
    is reshaped for broadcasting compatibility. The resulting tensors contain rotary embeddings and are
    returned as real tensors.

    Args:
        xq (torch.Tensor): Query tensor to apply rotary embeddings.
        xk (torch.Tensor): Key tensor to apply rotary embeddings.
        freqs_cis (torch.Tensor): Precomputed frequency tensor for complex exponentials.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Tuple of modified query tensor and key tensor with rotary embeddings.
    """
    xq_ = torch.view_as_complex(xq.float().reshape(*xq.shape[:-1], -1, 2))
    xk_ = torch.view_as_complex(xk.float().reshape(*xk.shape[:-1], -1, 2))
    freqs_cis = reshape_for_broadcast(freqs_cis, xq_)
    xq_out = torch.view_as_real(xq_ * freqs_cis).flatten(3)
    xk_out = torch.view_as_real(xk_ * freqs_cis).flatten(3)
    return xq_out.type_as(xq), xk_out.type_as(xk)


def repeat_kv(x: torch.Tensor, n_rep: int) -> torch.Tensor:
    """torch.repeat_interleave(x, dim=2, repeats=n_rep)"""
    bs, slen, n_kv_heads, head_dim = x.shape
    if n_rep == 1:
        return x
    return (
        torch.unsqueeze(x, dim=3)
        .expand(bs, slen, n_kv_heads, n_rep, head_dim)
        .reshape(bs, slen, n_kv_heads * n_rep, head_dim)
    )


class Attention(nn.Module):
    """Multi-head attention module.

    Args:
        model_args (ModelArgs): Model configuration arguments.

    Attributes:
        n_kv_heads (int): Number of key and value heads.
        n_heads (int): Number of query heads.
        n_rep (int): Number of repetitions for local heads.
        head_dim (int): Dimension size of each attention head.
        wq (Linear): Linear transformation for queries.
        wk (Linear): Linear transformation for keys.
        wv (Linear): Linear transformation for values.
        wo (Linear): Linear transformation for output.
    """

    def __init__(self, model_args: ModelCfg):
        super().__init__()
        self.n_heads = model_args.n_heads
        self.n_kv_heads = (
            model_args.n_heads
            if model_args.n_kv_heads is None
            else model_args.n_kv_heads
        )
        self.n_rep = self.n_heads // self.n_kv_heads
        self.head_dim = model_args.dim // model_args.n_heads

        self.wq = nn.Linear(
            model_args.dim, model_args.n_heads * self.head_dim, bias=False
        )
        self.wk = nn.Linear(model_args.dim, self.n_kv_heads * self.head_dim, bias=False)
        self.wv = nn.Linear(model_args.dim, self.n_kv_heads * self.head_dim, bias=False)
        self.wo = nn.Linear(
            model_args.n_heads * self.head_dim, model_args.dim, bias=False
        )

    def init_weights(self, init_std: float):
        for linear in (self.wq, self.wk, self.wv):
            nn.init.trunc_normal_(linear.weight, mean=0.0, std=0.02)
        nn.init.trunc_normal_(self.wo.weight, mean=0.0, std=init_std)

    def forward(
        self,
        x: torch.Tensor,
        freqs_cis: torch.Tensor,
    ):
        """Forward pass of the attention module.

        Args:
            x (torch.Tensor): Input tensor.
            freqs_cis (torch.Tensor): Precomputed frequency tensor.

        Returns:
            torch.Tensor: Output tensor after attention.
        """
        bs, seqlen, _ = x.shape
        xq, xk, xv = self.wq(x), self.wk(x), self.wv(x)

        # Use -1 instead of `n_heads` (or `n_kv_heads`) to infer the actual
        # local heads from sizes of xq, xk, and xv as TP may have sharded them
        # after the above linear ops.
        xq = xq.view(bs, seqlen, -1, self.head_dim)
        xk = xk.view(bs, seqlen, -1, self.head_dim)
        xv = xv.view(bs, seqlen, -1, self.head_dim)

        xq, xk = apply_rotary_emb(xq, xk, freqs_cis=freqs_cis)

        # repeat k/v heads if n_kv_heads < n_heads
        keys = repeat_kv(xk, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)
        values = repeat_kv(xv, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)

        xq = xq.transpose(1, 2)  # (bs, n_local_heads, seqlen, head_dim)
        xk = keys.transpose(1, 2)  # (bs, n_local_heads, seqlen, head_dim)
        xv = values.transpose(1, 2)  # (bs, n_local_heads, seqlen, head_dim)

        # we use casual mask for training
        output = F.scaled_dot_product_attention(xq, xk, xv, is_causal=True)
        output = output.transpose(
            1, 2
        ).contiguous()  # (bs, seqlen, n_local_heads, head_dim)
        output = output.view(bs, seqlen, -1)
        return self.wo(output)


class FeedForward(nn.Module):
    """FeedForward module.

    Args:
        dim (int): Input dimension.
        hidden_dim (int): Hidden dimension of the feedforward layer.
        multiple_of (int): Value to ensure hidden dimension is a multiple of this value.
        ffn_dim_multiplier (Optional[float]): Custom multiplier for hidden dimension. Defaults to None.

    Attributes:
        w1 (Linear): Linear transformation for the first layer.
        w2 (Linear): Linear transformation for the second layer.
        w3 (Linear): Linear transformation for the third layer.
    """

    def __init__(
        self,
        dim: int,
        hidden_dim: int,
        multiple_of: int,
        ffn_dim_multiplier: Optional[float],
    ):
        super().__init__()
        hidden_dim = int(2 * hidden_dim / 3)
        # custom dim factor multiplier
        if ffn_dim_multiplier is not None:
            hidden_dim = int(ffn_dim_multiplier * hidden_dim)
        hidden_dim = multiple_of * ((hidden_dim + multiple_of - 1) // multiple_of)

        self.w1 = nn.Linear(dim, hidden_dim, bias=False)
        self.w2 = nn.Linear(hidden_dim, dim, bias=False)
        self.w3 = nn.Linear(dim, hidden_dim, bias=False)

    def forward(self, x):
        return self.w2(F.silu(self.w1(x)) * self.w3(x))

    def init_weights(self, init_std: float):
        nn.init.trunc_normal_(self.w1.weight, mean=0.0, std=0.02)
        for linear in (self.w2, self.w3):
            nn.init.trunc_normal_(linear.weight, mean=0.0, std=init_std)


class TransformerBlock(nn.Module):
    """TransformerBlock Module.

    Args:
        layer_id (int): Identifier for the layer.
        model_args (ModelArgs): Model configuration arguments.

    Attributes:
        n_heads (int): Number of attention heads.
        dim (int): Dimension size of the model.
        head_dim (int): Dimension size of each attention head.
        attention (Attention): Attention module.
        feed_forward (FeedForward): FeedForward module.
        layer_id (int): Identifier for the layer.
        attention_norm (RMSNorm): Layer normalization for attention output.
        ffn_norm (RMSNorm): Layer normalization for feedforward output.
    """

    def __init__(self, layer_id: int, model_args: ModelCfg):
        super().__init__()
        self.n_heads = model_args.n_heads
        self.dim = model_args.dim
        self.attention = Attention(model_args)
        self.feed_forward = FeedForward(
            dim=model_args.dim,
            hidden_dim=4 * model_args.dim,
            multiple_of=model_args.multiple_of,
            ffn_dim_multiplier=model_args.ffn_dim_multiplier,
        )
        self.layer_id = layer_id
        self.num_layers = model_args.n_layers

        self.attention_norm = RMSNorm(dim=model_args.dim, eps=model_args.norm_eps)
        self.ffn_norm = RMSNorm(dim=model_args.dim, eps=model_args.norm_eps)

        if model_args.depth_init:
            self.weight_init_std = 0.02 / (2 * (self.layer_id + 1)) ** 0.5
        else:
            self.weight_init_std = 0.02 / (2 * self.num_layers) ** 0.5

    def forward(
        self,
        x: torch.Tensor,
        freqs_cis: torch.Tensor,
    ):
        """Perform a forward pass through the TransformerBlock.

        Args:
            x (torch.Tensor): Input tensor.
            freqs_cis (torch.Tensor): Precomputed cosine and sine frequencies.

        Returns:
            torch.Tensor: Output tensor after applying attention and feedforward layers.
        """
        h = x + self.attention(self.attention_norm(x), freqs_cis)
        out = h + self.feed_forward(self.ffn_norm(h))
        return out

    def init_weights(self):
        for norm in (self.attention_norm, self.ffn_norm):
            norm.reset_parameters()
        self.attention.init_weights(self.weight_init_std)
        self.feed_forward.init_weights(self.weight_init_std)


class Transformer(nn.Module):
    """Transformer Module.

    Args:
        model_args (ModelArgs): Model configuration arguments.

    Attributes:
        model_args (ModelArgs): Model configuration arguments.
        vocab_size (int): Vocabulary size.
        n_layers (int): Number of layers in the model.
        tok_embeddings (ParallelEmbedding): Token embeddings.
        layers (torch.nn.ModuleList): List of Transformer blocks.
        norm (RMSNorm): Layer normalization for the model output.
        output (ColumnParallelLinear): Linear layer for final output.
        freqs_cis (torch.Tensor): Precomputed cosine and sine frequencies.
    """

    def __init__(self, model_args: ModelCfg):
        super().__init__()
        self.model_args = model_args
        self.vocab_size = model_args.vocab_size
        self.n_layers = model_args.n_layers

        self.tok_embeddings = nn.Embedding(model_args.vocab_size, model_args.dim)

        # TODO persistent should be set to false, since this buffer can be recomputed.
        # however, we set it to true for 2 reasons.  (1) due to pytorch/pytorch#123411,
        # compile or pipeline-tracer will not correctly handle non-persistent buffers,
        # so we need to fix that.  (2) if we initialize pipeline-parallel models from
        # a seed checkpoint rather than calling init_weights, we need freqs_cis to be
        # initialized by the checkpoint, or we need to add a separate initializer for
        # just the non-persistent buffers that is called after loading checkpoints.
        self.register_buffer("freqs_cis", self._precompute_freqs_cis(), persistent=True)

        self.layers = torch.nn.ModuleDict()
        for layer_id in range(model_args.n_layers):
            self.layers[str(layer_id)] = TransformerBlock(layer_id, model_args)

        self.norm = RMSNorm(dim=model_args.dim, eps=model_args.norm_eps)

        self.output = nn.Linear(model_args.dim, model_args.vocab_size, bias=False)
        self.init_weights()

    def reset_parameters(self):
        self.init_weights()

    def init_weights(self):
        """
        [Note: On ``init_weights`` vs. ``reset_parameters``]
        Modules may define ``reset_parameters`` to initialize parameter values.
        ``reset_parameters`` is meant to only initialize directly owned
        parameters/buffers, not those of their child modules, and it can be
        used to give the initial values for these tensors.
        Separately, users may want custom initialization for their modules,
        different from that in ``reset_parameters``. For this, we define
        ``init_weights``. We only call it in the constructor of this
        ``Transformer`` root module to avoid reinitializing tensors.
        """
        with torch.device(self.freqs_cis.device):
            self.freqs_cis = self._precompute_freqs_cis()
        if self.tok_embeddings is not None:
            nn.init.normal_(self.tok_embeddings.weight)
        for layer in self.layers.values():
            if layer is not None:
                layer.init_weights()
        if self.norm is not None:
            self.norm.reset_parameters()
        final_out_std = self.model_args.dim**-0.5
        cutoff_factor = 3
        if self.output is not None:
            nn.init.trunc_normal_(
                self.output.weight,
                mean=0.0,
                std=final_out_std,
                a=-cutoff_factor * final_out_std,
                b=cutoff_factor * final_out_std,
            )

    def _precompute_freqs_cis(self) -> torch.Tensor:
        return precompute_freqs_cis(
            self.model_args.dim // self.model_args.n_heads,
            # Need to compute until at least the max token limit for generation
            # (use 2x max sequence length to be safe)
            self.model_args.max_seq_len * 2,
            self.model_args.rope_theta,
        )

    def forward(self, tokens: torch.Tensor):
        """Perform a forward pass through the Transformer model.

        Args:
            tokens (torch.Tensor): Input token indices.

        Returns:
            torch.Tensor: Output logits after applying the Transformer model.
        """
        # passthrough for nonexistent layers, allows easy configuration of pipeline parallel stages
        h = self.tok_embeddings(tokens) if self.tok_embeddings else tokens

        for layer in self.layers.values():
            h = layer(h, self.freqs_cis)

        h = self.norm(h) if self.norm else h
        output = self.output(h).float() if self.output else h
        return output

    @classmethod
    def from_model_args(cls, model_args: ModelCfg) -> "Transformer":
        """Initialize a Transformer model from a ModelArgs object.

        Args:
            model_args (ModelArgs): Model configuration arguments.

        Returns:
            Transformer: Transformer model.
        """
        return cls(model_args)



import operator
from fts_examples.patching._patch_utils import _prepare_module_ctx, lwt_compare_version

# we ignore these for the entire file since we're using our global namespace trickeration to patch
# ruff: noqa: F821
# pyright: reportUndefinedVariable=false

if lwt_compare_version("torch", operator.ge, "2.5.0"):
    globals().update(_prepare_module_ctx('torch.distributed.tensor._ops._einsum_strategy', globals()))


    def gen_einsum_strategies(
        equation: str,
        mesh: DeviceMesh,
        *,
        linearity: bool = False,
    ) -> OpStrategy:
        """Generate a strategy list for the ops that follow einsum style notation."""
        # parse einop equation and extract dims
        input_dims, output_dim = EinsumDims.parse_equation(equation)
        edims = EinsumDims.parse_dims(input_dims, output_dim)

        all_mesh_dim_strategies = []

        # generate strategies for each mesh dim
        for mesh_dim in range(mesh.ndim):
            mesh_dim_strategies = []

            # placement list stores placements of [output, input1, input2, ...]
            # first we always have replicate all for inputs and output
            placement_list: List[Placement] = [Replicate()] * (len(input_dims) + 1)
            mesh_dim_strategies.append(placement_list)

            # if mesh.size(mesh_dim) <= 1:
            #     # only replicate strategy for mesh dim with size 1
            #     # TODO: see if this is valid for the submesh case
            #     continue

            # split batch dim
            for batch_dim in edims.batch_dims:
                output_batch_dim = output_dim.index(batch_dim)
                placement_list = [Shard(output_batch_dim)]
                for input_dim in input_dims:
                    input_batch_dim = input_dim.index(batch_dim)
                    placement_list.append(Shard(input_batch_dim))

                mesh_dim_strategies.append(placement_list)

            # split contracting dim
            for contracting_dim in edims.contracting_dims:
                placement_list = [Partial()]
                for input_dim in input_dims:
                    input_contracting_dim = input_dim.index(contracting_dim)
                    placement_list.append(Shard(input_contracting_dim))

                mesh_dim_strategies.append(placement_list)

            # split lhs free dim
            for lhs_dim in edims.lhs_out_only_dims:
                lhs_free_dim = output_dim.index(lhs_dim)
                # this means split the lhs input and output
                # i.e. S(0), R -> S(0)
                lhs_placement_list: List[Placement] = [
                    Shard(lhs_free_dim),
                    Shard(lhs_free_dim),
                    Replicate(),
                ]
                mesh_dim_strategies.append(lhs_placement_list)

            # split rhs free dim
            for rhs_dim in edims.rhs_out_only_dims:
                rhs_free_dim = output_dim.index(rhs_dim)
                rhs_placement_list: List[Placement] = [
                    Shard(rhs_free_dim),
                    Replicate(),
                    Shard(rhs_free_dim),
                ]
                mesh_dim_strategies.append(rhs_placement_list)

            # linearity strategy
            if linearity:
                linearity_placement_list: List[Placement] = [Partial()]
                for input_dim in input_dims:
                    linearity_placement_list.append(Partial())
                mesh_dim_strategies.append(linearity_placement_list)

            all_mesh_dim_strategies.append(mesh_dim_strategies)

        # generate strategies for entire mesh
        strategy_combs = itertools.product(*all_mesh_dim_strategies)

        # TODO: filter out invalid strategies, at this point we generate
        # all possible strategies without considering the whether the tensor
        # dim could be sharded or not, we would need to filter out invalid
        # strategies base on the actual tensor shape
        # (i.e. for Shard, tensor dim size must > mesh size)
        all_strategies = []
        for strategy_comb in strategy_combs:
            spec_list = []
            for specs in zip(*strategy_comb):
                spec_list.append(DTensorSpec(mesh, tuple(specs)))
            strat = PlacementStrategy(output_specs=spec_list[0], input_specs=spec_list[1:])
            all_strategies.append(strat)

        return OpStrategy(all_strategies)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
MemProfiler Configuration Dataclasses
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module defines the configuration dataclasses for the MemProfiler.
"""
from typing import Optional, Callable, List, Set, Union
from dataclasses import dataclass, field, fields
from pathlib import Path

import torch
from lightning.fabric.utilities import rank_zero_warn

# conditionally import indirectly to avoid duplicating import logic in several different modules
from finetuning_scheduler.strategy_adapters._mp_imports import _TORCH_GREATER_EQUAL_2_5


@dataclass
class MemProfilerHooks:
    pre_forward_hooks: List[Union[str, Callable]] = \
        field(default_factory=lambda: ['fts_examples.profiling.npp_hooks._hook_npp_pre_forward'])
    post_forward_hooks: List[Union[str, Callable]] = \
        field(default_factory=lambda: ['fts_examples.profiling.npp_hooks._hook_npp_post_forward'])
    # the provided reset_state_hooks will be called with the model and the `save_hook_attrs` list
    reset_state_hooks: List[Union[str, Callable]] = \
        field(default_factory=lambda: ['fts_examples.profiling.npp_hooks._reset_memory_hooks_state'])

@dataclass
class MemProfilerFuncs: # can specify arbitrary list of `memprofilable` decorated function names
    # funcs that will be added to all memory collection types
    default: Set[str] = field(default_factory=lambda: {'training_step'})
    cpu: Set[str] = field(default_factory=set)
    cuda: Set[str] = field(default_factory=set)
    cuda_allocator_history: Set[str] = field(default_factory=set)
    fsdp: Set[str] = field(default_factory=set)

@dataclass
class MemProfilerSchedule:
    # keeping schedule simple as possibile for now, may expand to accommodate more flexible schedules in the future
    warmup_iters: int = 1
    max_iter: Optional[int] = None

@dataclass
class MemProfilerCfg:
    """
    Configuration dataclass for the MemProfiler.

    :param enabled: Whether to enable memory profiling.
    :param collect_funcs: A MemProfilerFuncs instance specifying the functions to collect per memory collection type.
    :param cuda_allocator_history: Whether to collect CUDA memory allocator history.
    :param track_fsdp_mem: Whether to collect FSDP memory statistics.
    :param fsdp_mem_track_module_depth: The depth of FSDP modules to track.
    :param fsdp_mem_tracker_tabulate: Whether to print FSDP memory statistics in a tabular format.
    :param fsdp_mem_tracker_units: The units to use for FSDP memory statistics.
    :param fsdp_mem_tracker_root_module: The root module to use for FSDP memory statistics.
    :param dump_memorystats_pickle: Whether to dump memory statistics to a pickle file.
    :param dump_memorystats_yaml: Whether to dump memory statistics to a yaml file.
    :param schedule: A MemProfilerSchedule instance specifying the schedule for memory collection.
    :param save_dir: The directory to save the memory statistics.
    :param enable_memory_hooks: Whether to enable memory hooks.
    :param enable_saved_tensors_hooks: Whether to enable saved tensors hooks.
    :param memory_hooks: A MemProfilerHooks instance specifying the memory hooks.
    :param saved_tensors_funcs: A list of saved tensors functions.
    :param save_hook_attrs: A list of module state attributes to save.
    :param retain_hooks_for_funcs: A set of functions to retain memory hooks for.
    """
    enabled: bool = False
    # specify funcs to collect per memory collection type, a default list to apply to all types or both composed
    collect_funcs: MemProfilerFuncs = field(default_factory=MemProfilerFuncs)
    cuda_allocator_history: bool = False
    track_fsdp_mem: bool = False
    fsdp_mem_track_module_depth: int = 2
    fsdp_mem_tracker_tabulate: bool = False
    fsdp_mem_tracker_units: str = "MiB"
    fsdp_mem_tracker_root_module: str = ""
    dump_memorystats_pickle: bool = False
    dump_memorystats_yaml: bool = True
    schedule: MemProfilerSchedule = field(default_factory=MemProfilerSchedule)
    save_dir: Optional[Union[str, Path]] = None
    enable_memory_hooks: bool = True
    enable_saved_tensors_hooks: bool = True
    memory_hooks: MemProfilerHooks = field(default_factory=MemProfilerHooks)
    # because it's frequently used for unpacking and to ensure this dataclass remains serializable, we allow
    # specification of 'identity_lambda' which will resolve to `lambda x: x`
    saved_tensors_funcs: List = field(default_factory=lambda: list(('fts_examples.profiling.npp_hooks._npp_hook',
                                                                    'identity_lambda')))
    # if you add custom hooks, make sure to add the desired module state attributes to save to `save_hook_attrs`
    save_hook_attrs: List = field(default_factory=lambda: ["rss_pre_forward", "rss_post_forward", "rss_diff",
                                                           "npp_pre_forward", "npp_post_forward", "npp_diff"])
    # since we cannot reliably ascertain when all MemProfilerFuncs will be executed, memory hooks will
    # only be removed once the funcs in this set have reached `max_iter`
    retain_hooks_for_funcs: Set[str] = field(default_factory=lambda: {'training_step'})

    def __post_init__(self) -> None:
        if not self.enabled:
            return
        if not torch.cuda.is_available() and any((self.collect_funcs.cuda_allocator_history, self.collect_funcs.cuda,
                                                  self.cuda_allocator_history)):
            rank_zero_warn("Disabling CUDA memory profiling functionality since no CUDA device detected.")
            self.collect_funcs.cuda, self.collect_funcs.cuda_allocator_history = set(), set()
            self.cuda_allocator_history = False
        if self.track_fsdp_mem and not _TORCH_GREATER_EQUAL_2_5:
            rank_zero_warn("Disabling FSDP memory profiling functionality since PyTorch version < 2.5.")
            self.track_fsdp_mem = False
        has_hooks = any(getattr(self.memory_hooks, ht.name) for ht in fields(self.memory_hooks))
        if not has_hooks:
            rank_zero_warn("MemProfilerCfg is configured to enable memory hooks but MemProfilerHooks does not have"
                           " any specified.")
        if self.schedule.max_iter is None:
            self.schedule.max_iter = self.schedule.warmup_iters + 1
        # compose all non-default func sets with the default set
        default_funcs = self.collect_funcs.default
        for k in self.collect_funcs.__dataclass_fields__.keys():
            if k != 'default':
                getattr(self.collect_funcs, k).update(default_funcs)



from lightning_utilities.core.imports import module_available

_HF_AVAILABLE = module_available("transformers") and module_available("datasets")
_SP_AVAILABLE = module_available("sentencepiece")

from fts_examples.patching.dep_patch_shim import _ACTIVE_PATCHES  # noqa: E402, F401



import os
if not os.getenv("SHOW_EXAMPLE_CPP_WARNS", "0") == "1":
    os.environ["TORCH_CPP_LOG_LEVEL"] = "ERROR"
import warnings

import torch
import torch.nn.functional as F
import lightning as L

from fts_examples.cli_experiment_utils import ExpHarness, FTSExperimentCLI, ExperimentCfg
from fts_examples.model_parallel.torchtitan_llama import ModelCfg, Transformer
from finetuning_scheduler.strategy_adapters._mp_imports import (
    _TORCH_GREATER_EQUAL_2_5, ColwiseParallel, Replicate, loss_parallel, PrepareModuleInput, RowwiseParallel,
    SequenceParallel, parallelize_module, DeviceMesh, Shard)

# Lightning ModelParallel still uses `torch.load` with `weights_only=False`
warnings.filterwarnings("ignore", ".*uses the default pickle.*")

# modified version of https://bit.ly/torchtitan_transformer_tp_plan
def apply_tp_plan(model: Transformer, device_mesh: DeviceMesh, loss_parallel: bool) -> Transformer:
    """Apply parallelism."""
    tp_mesh = device_mesh["tensor_parallel"]

    # 1. Parallelize the pre-transformer block layers
    non_transformerblock_tp_plan = {
        # Parallelize the embedding and shard its outputs (which are the first transformer block's inputs)
        "tok_embeddings": RowwiseParallel(input_layouts=Replicate(), output_layouts=Shard(1),),
        "norm": SequenceParallel(),  # Parallelize the root norm layer over the sequence dim
    }
    model = parallelize_module(model, tp_mesh, non_transformerblock_tp_plan)

    # 2. Parallelize each transformer block
    for transformer_block in model.layers.values():
        plan = {
            "attention": PrepareModuleInput(
            input_layouts=(Shard(1), None),
            desired_input_layouts=(Replicate(), None),
            ),
            "attention_norm": SequenceParallel(),
            "attention.wq": ColwiseParallel(),
            "attention.wk": ColwiseParallel(),
            "attention.wv": ColwiseParallel(),
            "attention.wo": RowwiseParallel(output_layouts=Shard(1)),
            "ffn_norm": SequenceParallel(),
            "feed_forward": PrepareModuleInput(
                input_layouts=(Shard(1),),
                desired_input_layouts=(Replicate(),),
            ),
            "feed_forward.w1": ColwiseParallel(),
            "feed_forward.w2": RowwiseParallel(output_layouts=Shard(1)),
            "feed_forward.w3": ColwiseParallel(),

        }

        parallelize_module(transformer_block, tp_mesh, plan)

    # 3. Parallelize the output layer
    output_parallelize_plan = (
        ColwiseParallel(
            input_layouts=Shard(1),
            output_layouts=Shard(-1) if loss_parallel else Replicate(),
            use_local_output=not loss_parallel,
        )
    )
    parallelize_module(model.output, tp_mesh, output_parallelize_plan)

    return model


class ModParallelExample(ExpHarness, L.LightningModule):
    def __init__(self, model_cfg: ModelCfg, exp_cfg: ExperimentCfg, *args, **kwargs):
        # the ExpHarness mixin just saves our hparams and collects our env info so our experiments can be nicely logged
        super().__init__(model_cfg=model_cfg, exp_cfg=exp_cfg, *args, **kwargs)
        with torch.device("meta"):
            self.model = Transformer(self.hparams.model_cfg)

    def on_train_start(self) -> None:
        self.model.init_weights()
        super().on_train_start()

    def configure_model(self):
        if self.device_mesh["tensor_parallel"].size() > 1:
            # User-defined function that applies the a given TP plan if desired
            apply_tp_plan(self.model, device_mesh=self.device_mesh, loss_parallel=self.hparams.exp_cfg.loss_parallel)

        # Note FTS will apply any module name/pattern-based `fsdp_plan` directives after any preceding Tensor
        # Parallel (like above) or explicit `fully_shard` directives here in `LightningModule.configure_model`. FTS will
        # only apply `fully_shard` to a specified module if it was not already applied to that module, so auto
        # `fsdp_plan` and manual `fully_shard` directives can be composed without conflict.

    def loss_fn(self, output: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if self.hparams.exp_cfg.loss_parallel:
            with loss_parallel():
                loss = F.cross_entropy(output.reshape(-1, output.size(-1)), target.reshape(-1))
        else:
            loss = F.cross_entropy(output.reshape(-1, output.size(-1)), target.reshape(-1))
        return loss

    def backward(self, *args, **kwargs):
        if self.hparams.exp_cfg.loss_parallel:
            with loss_parallel():
                super().backward(*args, **kwargs)
        else:
            super().backward(*args, **kwargs)


def cli_main() -> None:
    torch.set_float32_matmul_precision("high")
    assert torch.cuda.device_count() >= 2, "This example requires at least 2 GPUs"
    assert _TORCH_GREATER_EQUAL_2_5, "This example requires PyTorch 2.5 or higher"
    # every configuration of this example depends upon a shared set of defaults.
    default_config_file = os.path.join(os.path.dirname(__file__), "config", "defaults", "fts_mp_example_defaults.yaml")
    _ = FTSExperimentCLI(L.LightningModule, subclass_mode_model=True, save_config_kwargs={"overwrite": True},
        parser_kwargs={"fit": {"default_config_files": [default_config_file]}},
    )

if __name__ == "__main__":
    cli_main()



import operator
import sys
from typing import Callable
import importlib.metadata
from packaging.version import Version
from functools import lru_cache


@lru_cache
def lwt_compare_version(package: str, op: Callable, version: str, use_base_version: bool = True,
                        local_version: str = None) -> bool:
    try:
        pkg_version = Version(importlib.metadata.version(package))
    except (importlib.metadata.PackageNotFoundError):
        return False
    except TypeError:
        # possibly mocked by Sphinx so needs to return True to generate summaries
        return True
    if local_version:
        if not operator.eq(local_version, pkg_version.local):
            return False
    if use_base_version:
        pkg_version = Version(pkg_version.base_version)
    return op(pkg_version, Version(version))


def _prepare_module_ctx(module_path, orig_globals):
    _orig_file = orig_globals.pop('__file__')
    orig_globals.update(vars(sys.modules.get(module_path)))
    orig_globals['__file__'] = _orig_file
    return orig_globals



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List, Set, Iterator, Tuple
from types import resolve_bases

from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import _CHECKPOINT_PREFIX
import torch

if torch.distributed.is_available():
    from torch.distributed.fsdp.wrap import _Policy, CustomPolicy


class NameDrivenCustomPolicy(CustomPolicy):
    """An auto-wrapping policy extension that applies module name-based override directives on top of a given base
    ``auto_wrap_policy``.

    The composition of module name-based wrapping directives with a given ``auto_wrap_policy`` is
    achieved here by:
        1. Subclassing ``CustomPolicy``, initializing it with an object id-based module name mapping lambda to
            create a module name-driven policy with a base policy handle.
        2. Generating a base wrapping configuration dictionary by running the user's provided ``auto_wrap_policy``
        3. Updating the base configuration dictionary with any wrapping configuration generated by the module
            name-driven policy and returning the composed configuration.
    """

    def __init__(self, auto_wrap_policy_handle: _Policy, override_ids: List):
        """Compose the provided ``auto_wrap_policy`` with any provided override directives.

        Args:
            auto_wrap_policy_handle (Union[Callable, _Policy]): The user's base ``auto_wrap_policy``.
            override_ids (List): Object ids of the desired modules to wrap even if the provided ``auto_wrap_policy``
                otherwise would not dictate so.
        """
        super().__init__(lambda_fn=lambda m: id(m) in override_ids)
        self._base_awp = auto_wrap_policy_handle

    def _run_policy(
        self,
        root_module: torch.nn.Module,
        ignored_modules: Set[torch.nn.Module],
        root_kwargs: Dict[str, Any],
    ) -> Dict[torch.nn.Module, Dict[str, Any]]:
        target_module_to_kwargs = self._base_awp._run_policy(root_module, ignored_modules, root_kwargs)
        nb_policy_target_module_to_kwargs = super()._run_policy(root_module, ignored_modules, root_kwargs)
        target_module_to_kwargs.update(nb_policy_target_module_to_kwargs)
        return target_module_to_kwargs

NameDrivenPolicy = NameDrivenCustomPolicy


class NCACMixin(torch.nn.Module):

    # extracted from https://bit.ly/non_composable_ac_adapt
    def named_parameters(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Iterator[Tuple[str, torch.nn.Parameter]]:
        """Override :meth:`named_parameters()` to intercept parameter names.

        remove all occurrences of ``_CHECKPOINT_PREFIX``.
        """
        for param_name, param in super().named_parameters(*args, **kwargs):
            yield param_name.replace(_CHECKPOINT_PREFIX, ""), param


def _compose_ncac(m: torch.nn.Module) -> None:
    """Composes a given module with the NCACMixin to accommodate non-composable activation checkpoint (NCAC) API
    usage."""
    new_bases = resolve_bases((NCACMixin, m.__class__))
    patched_cls = type(f'{m.__class__.__name__}_NCACAdapted', new_bases, {})
    m.__class__ = patched_cls



# TODO: replace local version once Lightning version available
# from lightning.fabric.utilities.imports import _TORCH_GREATER_EQUAL_2_5
from lightning_utilities.core.imports import compare_version
import operator
_TORCH_GREATER_EQUAL_2_5 = compare_version("torch", operator.ge, "2.5.0", use_base_version=True)

# ruff: noqa: F401
# we require torch 2.5 or higher for composable distributed API support so until torch 2.5.0 is the minimum version,
# supported, we conditionally import indirectly to avoid duplicating import logic in several different modules
if _TORCH_GREATER_EQUAL_2_5:
    from torch.distributed.device_mesh import DeviceMesh
    from torch.distributed.tensor import DTensor, Replicate, Shard
    from torch.distributed.tensor.experimental import implicit_replication
    from torch.distributed.tensor.parallel import (ColwiseParallel, PrepareModuleInput, RowwiseParallel,
                                                   SequenceParallel, parallelize_module, loss_parallel)
    from torch.nn.attention import SDPBackend, sdpa_kernel
    from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import (checkpoint_wrapper, offload_wrapper,
                                                                             ActivationWrapper)
    from torch.distributed._composable import checkpoint
    from torch.distributed._composable.fsdp import CPUOffloadPolicy, FSDPModule, fully_shard
    from torch.distributed._tools.fsdp2_mem_tracker import FSDPMemTracker
else:
    for mp_obj in ["SDPBackend", "DeviceMesh", "DTensor", "Replicate", "Shard", "ColwiseParallel", "PrepareModuleInput",
                "RowwiseParallel", "SequenceParallel", "implicit_replication", "parallelize_module", "loss_parallel",
                "FSDPModule", "fully_shard", "checkpoint", "checkpoint_wrapper", "offload_wrapper", "ActivationWrapper",
                "CPUOffloadPolicy", "sdpa_kernel", "FSDPMemTracker"]:
        globals()[mp_obj] = object



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
MemProfiler
^^^^^^^^^^^

A powerful memory profiling utility that synthesizes numerous complementary profiling methods.
"""
import os
import pickle
from typing import Any, Dict, Optional, Tuple,  DefaultDict, Callable, List, Set
from dataclasses import dataclass, field, fields, asdict
from contextlib import redirect_stdout, contextmanager, ExitStack
from collections import defaultdict
from itertools import chain
from functools import wraps
from psutil import Process
from copy import deepcopy
from pathlib import Path

import yaml
import torch
from lightning.fabric.utilities.rank_zero import _get_rank
from lightning.fabric.utilities.cloud_io import get_filesystem
from lightning.pytorch.utilities.exceptions import MisconfigurationException

# conditionally import indirectly to avoid duplicating import logic in several different modules
from finetuning_scheduler.strategy_adapters._mp_imports import FSDPModule, FSDPMemTracker
from fts_examples.cfg_utils import resolve_funcs
from fts_examples.profiling.config import MemProfilerCfg


class MemProfiler:
    """MemProfiler is a powerful memory profiling utility that synthesizes numerous complementary profiling methods.

    The following profiling utilities are integrated and simultaneously configured:

    - ``FSDP2MemTracker``
    - `cuda memory snapshot and allocator history tracking <https://pytorch.org/docs/stable/torch_cuda_memory.html>`_
    - host-level memory tracking
    - custom memory hooks (e.g. for activation checkpoint memory tracking via ``saved_tensors_hooks`` etc.)

    .. warning::
        :class:`~fts_examples.profiling.memprofiler.MemProfiler` is in BETA and subject to change.
        The interface can bring breaking changes and new features with the next release of Finetuning Scheduler.
    """
    def __init__(self, *args, **kwargs) -> None:
        """
        The MemProfiler is a powerful memory profiling utility that synthesizes numerous complementary profiling
        methods.

        The following profiling utilities are integrated and simultaneously configured:

        - ``FSDP2MemTracker``
        - `cuda memory snapshot and allocator history
          tracking <https://pytorch.org/docs/stable/torch_cuda_memory.html>`_
        - host-level memory tracking
        - custom memory hooks (e.g. for activation checkpoint memory tracking via ``saved_tensors_hooks`` etc.)

        `See this example for usage guidance.
        <https://finetuning-scheduler.readthedocs.io/en/latest/profiling/memprofiler_profiling.html>`_

        Args:
            *args: Ignored positional arguments.
            **kwargs: Ignored keyword arguments.

        Attributes:
            memory_stats (DefaultDict): A dictionary of dictionaries containing memory statistics.
            fsdp_memory_stats (Dict): A dictionary containing FSDP memory statistics.
            rank (int): The rank of the current process.
            module (Any): The module being profiled.
            memprof_log_dir (str): The directory where memory profiles are saved.
            fsdp_mem_tracker (FSDPMemTracker): The FSDP memory tracker.
            saved_tensors_funcs (List[Callable]): A list of functions used to track saved tensors.
            _state (MemProfInternalState): The internal state of the MemProfiler.
        """
        super().__init__()
        self.memory_stats = defaultdict(dict)
        self.fsdp_memory_stats = {}
        self.rank = _get_rank() or 0
        self.module = None
        self.memprof_log_dir = None
        self.fsdp_mem_tracker = None
        self.saved_tensors_funcs: List[Callable] = []
        self._state = _MemProfInternalState()

    def connect(self, obj_ref: Any) -> None:
        """
        Connects the MemProfiler to the given target module.

        Args:
            obj_ref: The module to be profiled.
        """
        self.module = obj_ref
        self._state.curr_pid = Process(os.getpid())
        if self.memprofiler_cfg.enable_saved_tensors_hooks:
            self.saved_tensors_funcs = resolve_funcs(cfg_obj=self.memprofiler_cfg, func_type='saved_tensors_funcs')
        self._state.base_collect_func_set = self._all_base_collect_funcs()
        self.memprofiler_cfg.retain_hooks_for_funcs = self._state.base_collect_func_set  # conservatively wait for all

    def _all_base_collect_funcs(self) -> Set:
        funcs = self.memprofiler_cfg.collect_funcs
        return set(chain(*[funcs.cuda_allocator_history, funcs.cuda, funcs.cpu]))

    @property
    def memprofiler_cfg(self) -> MemProfilerCfg:
        return self.module.memprofiler_cfg

    @property
    def schedule(self) -> int:
        return self.module.memprofiler_cfg.schedule

    @property
    def fsdp_mem_tracker_root_module(self) -> Optional[FSDPModule]:  # type: ignore
        """
        If ``track_fsdp_mem`` is enabled, this is the root FSDP module used for FSDP2 memory tracking.
        The root module must have ``fully_shard`` applied for FSDP2 memory tracking.
        If ``track_fsdp_mem`` is disabled, this is ``None``.
        """
        if not self.memprofiler_cfg.track_fsdp_mem:
            return None
        try:
            fsdp_tracker_root = self.module.get_submodule(self.memprofiler_cfg.fsdp_mem_tracker_root_module)
            assert issubclass(type(fsdp_tracker_root), FSDPModule), \
                "MemProfiler requires the root FSDP module to have ``fully_shard`` applied for FSDP2 memory tracking."
            return fsdp_tracker_root
        except AttributeError:
            raise MisconfigurationException(
                "Could not find the specified ``fsdp_mem_tracker_root_module``: "
                f"{self.memprofiler_cfg.fsdp_mem_tracker_root_module}. MemProfiler needs to attach to a module that has"
                " ``fully_shard`` applied for FSDP2 memory tracking."
            )

    def remove_memprofiler_hooks(self) -> None:
        """Removes all hooks added by MemProfiler."""
        for handle_list in self._state.hook_handles.values():
            for handle in handle_list:
                handle.remove()

    def exec_reset_state_hooks(self) -> None:
        """
        Executes all hooks registered for resetting memory tracking state.

        These hooks are responsible for resetting state (e.g. saved tensor sizes, RSS, etc.) tracked by the hooks
        registered in :meth:`add_memprofiler_hooks`.
        """
        for hook in self._state.configured_hooks["reset_state_hooks"]:
            hook(self.module, self.memprofiler_cfg.save_hook_attrs)

    def maybe_init_fsdp_mem_tracker(self) -> None:
        """Initializes FSDP2 memory tracker if ``track_fsdp_mem`` is ``True``."""
        if self.memprofiler_cfg.track_fsdp_mem:
            self.fsdp_mem_tracker = FSDPMemTracker(self.fsdp_mem_tracker_root_module, self.module.trainer.optimizers[0])

    def add_memprofiler_hooks(self) -> None:
        """
        Adds all hooks registered for memory tracking.

        Currently, hooks are registered for the `pre_forward` and `post_forward` points. Hooks are added to the
        `modules` of the provided `module` and are responsible for tracking memory usage. Supported hooks are
        registered based on the configuration of `memory_hooks` in `memprofiler_cfg`.

        After registering all hooks, this method calls :meth:`exec_reset_state_hooks` to reset the state of all hooks.
        """
        # TODO: extend supported hook points (e.g. backward, etc.) and if/once supporting additional hook points,
        # use a hook_type to registration function mapping
        memory_hooks_cfg = self.memprofiler_cfg.memory_hooks
        for supported_hooks in fields(memory_hooks_cfg):
            if getattr(memory_hooks_cfg, supported_hooks.name):
                self._state.configured_hooks[supported_hooks.name] = resolve_funcs(cfg_obj=memory_hooks_cfg,
                                                                             func_type=supported_hooks.name)
        for module in self.module.modules():
            module.mem_info_handle = self._state.curr_pid.memory_info
            for hook_func in self._state.configured_hooks["pre_forward_hooks"]:
                self._state.hook_handles[hook_func].append(module.register_forward_pre_hook(hook_func))
            for hook_func in self._state.configured_hooks["post_forward_hooks"]:
                self._state.hook_handles[hook_func].append(module.register_forward_hook(hook_func))
        self.exec_reset_state_hooks()

    def init_memprof_log_dir(self) -> None:
        """
        Initializes the directory where MemProfiler will save profiling artifacts.

        The directory is determined by the following logic:

        1. If ``memprofiler_cfg.save_dir`` is set, use that directory.
        2. Otherwise, use a directory named ``memprofiler`` in the current
           Lightning ``log_dir``.
        """
        self.memprof_log_dir = self.memprofiler_cfg.save_dir or Path(self.module._trainer.log_dir) / "memprofiler"
        self.memprof_log_dir = Path(self.memprof_log_dir)  # ensure the dir is a Path
        self.memprof_log_dir.mkdir(exist_ok=True, parents=True)

    def cuda_allocator_history_snap(self, snap_key: Tuple) -> None:
        """
        Dumps a snapshot of the CUDA memory allocator history.

        The snapshot is saved to a file named ``cuda_alloc_rank_<snap_key>.pickle`` in the directory specified by
        ``memprof_log_dir``.

        Args:
            snap_key: A tuple used to identify the snapshot.
        """
        cuda_snapshot_file = (self.memprof_log_dir / f"cuda_alloc_rank_{snap_key}.pickle")
        torch.cuda.memory._dump_snapshot(cuda_snapshot_file)

    def done(self, iter_idx: int) -> bool:
        """
        Checks if the profiling is done.

        Args:
            iter_idx: The current iteration index.

        Returns:
            bool: Whether the profiling is done.
        """
        return self.schedule.max_iter and iter_idx > self.schedule.max_iter

    def _process_hooks(self, snap_key) -> None:
        if self.memprofiler_cfg.enable_memory_hooks:
            if len(self._state.hook_handles) == 0:
                self.add_memprofiler_hooks()
            collected = {attr: getattr(self.module, attr, None) for attr in self.memprofiler_cfg.save_hook_attrs}
            self.memory_stats[snap_key].update(collected)

    def _collect_snap(self, snap_key, reset_mem_hooks: bool = False) -> None:
        """
        Collects a snapshot of the current memory statistics.

        Args:
            snap_key: A tuple of strings identifying the snapshot.
            reset_mem_hooks: Whether to reset the memory hooks after collecting the snapshot.

        Notes:
            The snapshot is saved to `memory_stats` with the key `snap_key`.
        """
        _, fn_name, *_ = snap_key
        snap_key = ".".join(map(str, snap_key))
        mem_cfg = self.memprofiler_cfg
        self._process_hooks(snap_key)
        if fn_name in mem_cfg.collect_funcs.cpu:
            mem = self._state.curr_pid.memory_info()
            self.memory_stats[snap_key].update({"rss": mem.rss, "vms": mem.vms})
        if fn_name in mem_cfg.collect_funcs.cuda:
            self.memory_stats[snap_key].update(torch.cuda.memory_stats())
        if fn_name in mem_cfg.collect_funcs.cuda_allocator_history and mem_cfg.cuda_allocator_history:
            self.cuda_allocator_history_snap(snap_key)
        if mem_cfg.enable_memory_hooks and reset_mem_hooks:
            self.exec_reset_state_hooks()

    @property
    def _should_remove_hooks(self) -> bool:
        return all(func in self._state.done_prof_funcs for func in self.memprofiler_cfg.retain_hooks_for_funcs)

    def teardown_prof(self, fn_name: str, iter_ctx: str) -> None:
        """
        Tears down profiling state for a function.

        This method is responsible for:

        1. Disabling profiling for the given function and iteration context.
        2. Marking the function as done if it is not being profiled in either the "start" or "end" iteration contexts.
        3. Removing all hooks registered by MemProfiler if the function is marked as done and
           ``retain_hooks_for_funcs`` is ``True``.

        Args:
            fn_name (str): The name of the function to tear down profiling state for.
            iter_ctx (str): The iteration context of the function to tear down profiling state for.
        """
        self._state.can_collect[(fn_name, iter_ctx)] = False
        if not any(self._state.can_collect[(fn_name, iter_ctx)] for iter_ctx in ["start", "end"]):
            self._state.done_prof_funcs.append(fn_name)
        if self.memprofiler_cfg.retain_hooks_for_funcs and self._should_remove_hooks:
            self.remove_memprofiler_hooks()
            self.memprofiler_cfg.enable_memory_hooks = False

    def gen_snap_keys(self, fn_name: str, iter_ctx: str, iter_idx: Optional[int] = None) -> Tuple[int, int, Tuple]:
        """
        Generates the MemProfiler snapshot key for a given function and iteration context.

        Args:
            fn_name (str): The name of the function to generate a snapshot key for.
            iter_ctx (str): The iteration context of the function to generate a snapshot key for.
            iter_idx (Optional[int]): The iteration index to use in the snapshot key. If ``None``, the current iteration
                index will be used.

        Returns:
            A tuple of (iteration index, snapshot key). The snapshot key is a tuple of (rank, function name, iteration
            index, iteration context).

        Note:
            The snapshot key format is rank.fn_name.iter_idx.iter_ctx.
        """
        # NOTE [Memprofiler Key Format]
        # snap key format is rank.fn_name.iter_idx.iter_ctx
        # e.g. 0.training_step.0.end keys hook output for the end of `training_step` 0 (first iteration), for rank 0
        # 0.training_step.2.start keys mem stats for the start of `training_step` 2, for rank 0
        if iter_idx is None:
            iter_idx = self._state.snap_indices[(fn_name, iter_ctx)]
        return iter_idx, (self.rank, fn_name, iter_idx, iter_ctx)

    def update_collect_state(self, fn_name: str, iter_ctx: str, iter_idx: Optional[int] = None) -> None:
        """
        Updates the MemProfiler state for a given function and iteration context.

        Args:
            fn_name (str): The name of the function to update the state for.
            iter_ctx (str): The iteration context of the function to update the state for.
            iter_idx (int, optional): The iteration index to use when generating the snapshot key. If ``None`` is
                provided, the current iteration index will be used.
        """
        self._state.maybe_init_iter_state(fn_name, iter_ctx)
        if self._state.can_collect[(fn_name, iter_ctx)] and not self._state._iter_incremented[(fn_name, iter_ctx)]:
            self._state.snap_indices[(fn_name, iter_ctx)] += 1
            self._state._iter_incremented[(fn_name, iter_ctx)] = True
            curr = self._state
            curr._iter_idx, curr._snap_key = self.gen_snap_keys(fn_name, iter_ctx, iter_idx)

    def snap(self, fn_name: str, iter_ctx: str, reset_mem_hooks: bool = False) -> None:
        """
        Collects a memory snapshot for the given function and iteration context.

        If the current iteration index is greater than the number of warmup iterations,
        then this function will check if the iteration index is within the profiling
        range. If it is, then a memory snapshot will be collected. If not, the memory
        profiler state for the given function and iteration context will be torn down.

        Args:
            fn_name (str): The name of the function to collect a memory snapshot for.
            iter_ctx (str): The iteration context of the function to collect a memory snapshot for.
            reset_mem_hooks (bool, optional): Whether to reset the memory hooks after collecting the snapshot.
                Defaults to False.
        """
        if self._state._iter_idx > self.schedule.warmup_iters:
            if not self.done(self._state._iter_idx):
                self._collect_snap(self._state._snap_key, reset_mem_hooks)
            else:
                self.teardown_prof(fn_name, iter_ctx)

    def dump_memory_stats(self) -> None:
        """
        Dumps the collected memory statistics to a pickle file and/or a yaml file.

        If `dump_memorystats_pickle` is True, the memory statistics will be saved to a pickle file
        at `memprof_log_dir / f"rank_{self.rank}_memory_stats.pickle"`.

        If `dump_memorystats_yaml` is True, the memory statistics will be saved to a yaml file
        at `memprof_log_dir / f"rank_{self.rank}_memory_stats.yaml"`.

        If `track_fsdp_mem` is True, the collected FSDP memory statistics will be saved to a yaml file
        at `memprof_log_dir / f"rank_{self.rank}_fsdp_memory_stats.yaml"`.
        """
        if self.memory_stats:
            if self.memprofiler_cfg.dump_memorystats_pickle:
                filename = self.memprof_log_dir / f"rank_{self.rank}_memory_stats.pickle"
                with open(filename, "wb") as f:
                    pickle.dump(self.memory_stats, f)
            if self.memprofiler_cfg.dump_memorystats_yaml:
                filename = self.memprof_log_dir / f"rank_{self.rank}_memory_stats.yaml"
                with open(filename, "w", newline="") as f:
                    yaml.dump(self.memory_stats, f)
        if self.fsdp_memory_stats:
            self.save_fsdp_mem_reports()

    def save_fsdp_mem_reports(self) -> None:
        """
        Saves the collected FSDP memory statistics to a yaml file.

        For each function name and iteration in the collected FSDP memory statistics, a log file is created
        at `memprof_log_dir / f"fsdp_mem_rank_{self.rank}_{fn_name}_{iter}.log"` containing the FSDP memory
        statistics in both tabular and modulewise formats.

        The modulewise format is displayed up to the specified `fsdp_mem_track_module_depth`, and the units
        for the memory statistics are specified by `fsdp_mem_tracker_units`.

        If `fsdp_mem_tracker_tabulate` is True, the FSDP memory statistics will be displayed in a tabular format.
        Otherwise, a plain text format will be used.

        :return: None
        """
        for (fn_name, iter) in self.fsdp_memory_stats.keys():
            self.fsdp_mem_tracker.memory_tracking = self.fsdp_memory_stats[(fn_name, iter)]
            target_fsdp_depth = self.memprofiler_cfg.fsdp_mem_track_module_depth
            tabulate_mode = self.memprofiler_cfg.fsdp_mem_tracker_tabulate
            report_units = self.memprofiler_cfg.fsdp_mem_tracker_units
            memory_report = self.memprof_log_dir / f"fsdp_mem_rank_{self.rank}_{fn_name}_{iter}.log"
            fs = get_filesystem(memory_report)
            with fs.open(memory_report, "w", newline="") as fp:
                with redirect_stdout(fp):
                    fp.write(f"Iteration {iter} {fn_name} Peak FSDP memory stats: {os.linesep}")
                    fp.write("-" * 80 + os.linesep)
                    self.fsdp_mem_tracker.display_snapshot("peak", tabulate=tabulate_mode)
                    fp.write("-" * 80 + os.linesep)
                    fp.write(f"Iteration {iter} {fn_name} Modulewise FSDP memory stats: {os.linesep}")
                    fp.write("-" * 80 + os.linesep)
                    self.fsdp_mem_tracker.display_modulewise_snapshots(depth=target_fsdp_depth, units=report_units,
                                                                       tabulate=tabulate_mode)
                    fp.write("-" * 80 + os.linesep)

    def _in_collect_range(self) -> bool:
        return self._state._iter_idx > self.schedule.warmup_iters and self._state._iter_idx <= self.schedule.max_iter

    def _collecting(self, fn_name: str, iter_ctx: str = 'start') -> bool:
        return self._state.can_collect[(fn_name, iter_ctx)] and self._in_collect_range()

    def _should_collect_fsdp(self, fn_name: str) -> bool:
        return all((self.fsdp_mem_tracker, fn_name in self.memprofiler_cfg.collect_funcs.fsdp,
                    self._collecting(fn_name)))

    def _should_collect_saved_tensors(self, fn_name: str) -> bool:
        return self.memprofiler_cfg.enable_saved_tensors_hooks and self._collecting(fn_name)

    @contextmanager
    @staticmethod
    def memprofile_snap_ctx(memprofiler, fn_name: str):
        """
        Context manager for taking memory snapshots at the start and end of a method.

        This context manager takes care of calling the `snap` method of the `MemProfiler` instance
        at the start and end of the context. It also handles the case where we are only collecting
        FSDP memory statistics.

        Args:
            memprofiler (MemProfiler): The MemProfiler instance.
            fn_name (str): The name of the function being profiled.

        Yields:
            None
        """
        try:
            # no need to execute snap logic if we are only collecting fsdp memory stats
            if fn_name not in memprofiler._state.base_collect_func_set:
                yield
            else:
                memprofiler.update_collect_state(fn_name=fn_name, iter_ctx="start")
                memprofiler.snap(iter_ctx="start", fn_name=fn_name)
                if memprofiler._should_collect_saved_tensors(fn_name):
                    with torch.autograd.graph.saved_tensors_hooks(*memprofiler.saved_tensors_funcs):
                        yield
                else:
                    yield
        finally:
            if fn_name in memprofiler._state.base_collect_func_set:
                memprofiler.update_collect_state(fn_name=fn_name, iter_ctx="end")
                memprofiler.snap(iter_ctx="end", fn_name=fn_name, reset_mem_hooks=True)

    @contextmanager
    @staticmethod
    def memprofile_fsdp_ctx(memprofiler, fn_name: str, track_inputs_target: Optional[Tuple] = None):
        """Sets the FSDP memory tracker context manager if ``fsdp_mem_tracker_enabled`` is ``True``.

        This context manager takes care of calling the `update_collect_state` and `reset_mod_stats`
        methods of the `MemProfiler` instance as well as the `track_inputs` and `reset` methods of
        the `FSDPMemTracker` instance.

        Args:
            memprofiler (MemProfiler): The MemProfiler instance.
            fn_name (str): The name of the function being profiled.
            track_inputs_target (Optional[Tuple]): The FSDP inputs to track. If ``None``, no inputs will be tracked.

        Yields:
            None
        """
        fsdp_state_collected = False
        try:
            memprofiler.update_collect_state(fn_name=fn_name, iter_ctx="start")
            if track_inputs_target is not None:
                memprofiler.fsdp_mem_tracker.track_inputs(track_inputs_target)
            if not memprofiler._should_collect_fsdp(fn_name):
                yield
            else:
                fsdp_state_collected = True
                with memprofiler.fsdp_mem_tracker:
                    yield
        finally:
            memprofiler.update_collect_state(fn_name=fn_name, iter_ctx="end")
            if fsdp_state_collected:
                memprofiler.fsdp_memory_stats[(fn_name, memprofiler._state._iter_idx)] = \
                    deepcopy(memprofiler.fsdp_mem_tracker.memory_tracking)
                memprofiler.fsdp_mem_tracker.reset_mod_stats()

    @contextmanager
    @staticmethod
    def memprofile_meta_ctx(memprofiler, fn_name: str):
        """Context manager to set metadata for MemProfiler invocations.

        This context manager always sets the iteration state for the given function name
        to enable orchestration of relevant profiling invocations.

        Args:
            memprofiler (MemProfiler): The MemProfiler instance.
            fn_name (str): The name of the function being profiled.

        Yields:
            None
        """
        try:
            memprofiler._state.maybe_init_iter_state(fn_name, 'start')
            yield
        finally:
            memprofiler._state.reset_iter_state(fn_name, 'end')

    @staticmethod
    def memprofilable(func):
        """Decorate a function to enable use with MemProfiler.

        This decorator is used to enable profiling a method with MemProfiler.  If the
        method is decorated with this decorator and the object has an enabled MemProfiler
        instance, the method will be profiled using the MemProfiler instance.

        Args:
            func: The function to decorate.

        Returns:
            The decorated function.
        """
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.memprofiler:
                return func(self, *args, **kwargs)
            fn_name = func.__name__
            with MemProfiler.memprofile_ctx(self.memprofiler, fn_name=fn_name):
                return func(self, *args, **kwargs)
        return wrapper

    @contextmanager
    @staticmethod
    def memprofile_ctx(memprofiler, fn_name: str):
        """Context manager to orchestrate profiling for a given function.

        This context manager is responsible for calling the relevant subordinate context managers
        that manage profiling for a given function. The context managers invoked are:

        1. `memprofile_meta_ctx`: Sets metadata for the profiling invocation.
        2. `memprofile_fsdp_ctx`: Collects FSDP memory statistics.
        3. `memprofile_snap_ctx`: Collects memory snapshots at the start and end of a method.

        Args:
            memprofiler (MemProfiler): The MemProfiler instance.
            fn_name (str): The name of the function being profiled.

        Yields:
            None
        """
        ctx_managers = []
        ctx_managers.append(MemProfiler.memprofile_meta_ctx(memprofiler, fn_name=fn_name))
        ctx_managers.append(MemProfiler.memprofile_fsdp_ctx(memprofiler, fn_name=fn_name))
        ctx_managers.append(MemProfiler.memprofile_snap_ctx(memprofiler, fn_name=fn_name))
        with ExitStack() as stack:
            for ctx_manager in ctx_managers:
                stack.enter_context(ctx_manager)
            try:
                yield
            finally:
                stack.close()

def _memprofiler_cfg_mapping_representer(dumper, data):
    return dumper.represent_mapping('tag:yaml.org,2002:map', asdict(data))

yaml.SafeDumper.add_representer(MemProfilerCfg, _memprofiler_cfg_mapping_representer)


@dataclass
class _MemProfInternalState:
    can_collect: Dict[str, bool] = field(default_factory=dict)
    curr_pid: Optional[Process] = None
    snap_indices: Dict[str, int] = field(default_factory=dict)
    configured_hooks: Dict[str, Any] = field(default_factory=dict)
    hook_handles: DefaultDict[str, List[Any]] = field(default_factory=lambda: defaultdict(list))
    done_prof_funcs: List[str] = field(default_factory=list)
    base_collect_func_set: Optional[Set] = None
    _iter_idx: Optional[int] = None
    _snap_key: Optional[Tuple] = None
    _iter_incremented: Dict[str, int] = field(default_factory=dict)

    def maybe_init_iter_state(self, fn_name: str, iter_ctx: str) -> None:
        if not self.snap_indices.get((fn_name, iter_ctx), None):
            self._iter_incremented[(fn_name, iter_ctx)] = False
            self.snap_indices[(fn_name, iter_ctx)] = 0
            self.can_collect[(fn_name, iter_ctx)] = True

    def reset_iter_state(self, fn_name: str, iter_ctx: str) -> None:
        self._snap_key = None
        for (fn_name, iter_ctx) in self._iter_incremented.keys():
            self._iter_incremented[(fn_name, iter_ctx)] = False



import importlib
import yaml
from typing import Dict, List, Any, Union
from dataclasses import dataclass, field, asdict, fields

from transformers import PretrainedConfig
from lightning.pytorch.utilities.exceptions import MisconfigurationException


@dataclass
class ExperimentCfg:
    loss_parallel: bool = True
    experiment_tag: str = 'default'
    log_env_details: bool = True
    batch_size: int = 4
    num_workers: int = 2
    dataset_length: int = 32

@dataclass
class OptimizerCfg:
    class_fqn: Union[str, None] = None
    args: Dict = field(default_factory=dict)

@dataclass
class LRSchedulerCfg:
    class_fqn: Union[str, None] = None
    args: Dict = field(default_factory=dict)

@dataclass
class LightningLRSCfg:
    name: str = 'default'
    interval: str = 'epoch'
    frequency: int = 1

    def __post_init__(self):
        self._overridden = _is_overridden(self)

def optimizer_cfg_mapping_representer(dumper, data):
    return dumper.represent_mapping('tag:yaml.org,2002:map', asdict(data))

def lr_scheduler_cfg_mapping_representer(dumper, data):
    return dumper.represent_mapping('tag:yaml.org,2002:map', asdict(data))

def pretrained_cfg_mapping_representer(dumper, data):
    return dumper.represent_mapping('tag:yaml.org,2002:map', data.to_dict())

# Register all custom representers to both base dumper classes
representers = {
    PretrainedConfig: pretrained_cfg_mapping_representer,
    OptimizerCfg: optimizer_cfg_mapping_representer,
    LRSchedulerCfg: lr_scheduler_cfg_mapping_representer
}

for dumper_cls in [yaml.Dumper, yaml.SafeDumper]:
    for cls, representer in representers.items():
        dumper_cls.add_multi_representer(cls, representer)

def _is_overridden(dataclass_instance) -> bool:
    is_overridden = False
    for f in fields(dataclass_instance.__class__):
        if getattr(dataclass_instance, f.name) != f.default:
            is_overridden = True
            break
    return is_overridden

def resolve_funcs(cfg_obj: Any, func_type: str) -> List:
    resolved_funcs = []
    funcs_to_resolve = getattr(cfg_obj, func_type)
    if not isinstance(funcs_to_resolve, list):
        funcs_to_resolve = [funcs_to_resolve]
    for func_or_qualname in funcs_to_resolve:
        if callable(func_or_qualname):
            resolved_funcs.append(func_or_qualname)  # TODO: inspect if signature is appropriate for custom hooks
        elif func_or_qualname == 'identity_lambda':  # special case often used e.g. w/ saved_tensors_hooks
            resolved_funcs.append(lambda x: x)
        else:
            try:
                module, func = func_or_qualname.rsplit(".", 1)
                mod = importlib.import_module(module)
                resolved_func = getattr(mod, func, None)
                if callable(resolved_func):
                    resolved_funcs.append(resolved_func)
                else:
                    raise MisconfigurationException(f"Custom function {func} from module {module} is not callable!")
            except (AttributeError, ImportError) as e:
                err_msg = f"Unable to import and resolve specified function {func} from module {module}: {e}"
                raise MisconfigurationException(err_msg)
    return resolved_funcs



import os
from typing import Any, Callable, Dict, Optional, Tuple, KeysView, Union
from copy import deepcopy, copy
from logging import DEBUG
from pathlib import Path
from unittest import mock
from functools import partialmethod
from dataclasses import dataclass, field
from itertools import chain

import pytest
import torch
import torch.nn.functional as F
from torch import nn, Tensor
from torch.utils.data import DataLoader
from lightning.pytorch import seed_everything, Trainer
from lightning.pytorch.plugins.precision.fsdp import FSDPPrecision
from lightning.pytorch.strategies import ModelParallelStrategy
from lightning.pytorch.utilities.exceptions import MisconfigurationException
from lightning.pytorch.utilities.types import STEP_OUTPUT

from finetuning_scheduler import FinetuningScheduler, FTSCheckpoint, FTSEarlyStopping
from finetuning_scheduler.strategy_adapters import ModelParallelStrategyAdapter
from finetuning_scheduler.strategy_adapters.model_parallel import ActCkptEnum
# conditionally import indirectly to avoid duplicating import logic in several different modules
from finetuning_scheduler.strategy_adapters._mp_imports import (SDPBackend, DeviceMesh, DTensor, Replicate, Shard,
    ColwiseParallel, PrepareModuleInput, RowwiseParallel, SequenceParallel, implicit_replication, parallelize_module,
    loss_parallel, FSDPModule, fully_shard, _TORCH_GREATER_EQUAL_2_5)
from tests.helpers.boring_models import FTSToyTransformer, TestModelArgs, FTSWikiText2
from tests.helpers.common import (ExpectedResults, fts_check_warns, pytest_param_factory, get_fts,
                                  default_fts_sanity_chk, DeviceMeshSummary)
from tests.model_parallel_expected_paths import (path_fsdp_autocm, path_tp_fsdp_autocm,
                                                 path_fsdp, path_tp)
from tests.helpers.runif import RunIf
from tests.test_finetuning_scheduler_callback import (
    EXPECTED_WARNS,
    FinetuningSchedulerBoringModel,
    TestFinetuningScheduler,
    get_sched_fixture_tmpdir,
)


FTS_GLOBAL_STATE_LOG_MODE = os.environ.get("FTS_GLOBAL_STATE_LOG_MODE", "0") == "1"
MODEL_PARALLEL_BASE_WARNS = copy(EXPECTED_WARNS)
additional_model_parallel_warns = [
    "The number of training batches",  # minimizing cost of training for these tests
    "when logging on epoch level in distributed",  # validating FTS handling in this scenario
    "You are using `torch.load` with `weights_only=False`",  # known required w/ Lightning 2.4
    "of Tensor.is_pinned",
    "of Tensor.pin_memory",
]
MODEL_PARALLEL_BASE_WARNS.extend(additional_model_parallel_warns)
MODEL_PARALLEL_DYNAMO_EXPECTED_WARNS = []


class FSDPStateInspectMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.expected_fsdp2_modules = None

    def on_train_start(self, trainer, pl_module) -> None:
        if self._uses_fsdp2:
            self.expected_fsdp2_modules = set(
                chain(getattr(self.pl_module, 'cm_fsdp_plan', {}).keys(),
                      getattr(self.strategy_adapter, "fsdp_plan", {}).keys()))

    def on_train_epoch_end(self, trainer, pl_module) -> None:
        if self._uses_fsdp2:
            self._assert_fsdp_state()

    def on_val_epoch_end(self, trainer, pl_module) -> None:
        if self._uses_fsdp2:
            self._assert_fsdp_state()

    @property
    def _uses_fsdp_name_based_plan(self) -> bool:
        return any(self.strategy_adapter_cfg.get(k) for k in ("fsdp_plan", "fsdp_default_kwargs"))

    @property
    def _uses_fsdp2(self) -> bool:
        return any((self.pl_module.cm_fsdp_plan, self._uses_fsdp_name_based_plan))

    def _assert_fsdp_state(self) -> None:
        precision = None
        if self.pl_module.precision_key == "auto_16":
            assert isinstance(self.pl_module.trainer.strategy.precision_plugin, FSDPPrecision)
            precision = torch.float16 if self.pl_module.trainer.precision == "16-true" else torch.bfloat16
        for n, m in self.pl_module.named_modules():
            if n in self.expected_fsdp2_modules:
                self._inspect_composable_fsdp_state(m, precision)
            else:
                assert not issubclass(type(m), FSDPModule)  # orig module should not be composed with FSDPModule

    def _inspect_composable_fsdp_state(self, m: nn.Module, precision: Optional[torch.dtype] = None) -> None:
            assert issubclass(type(m), FSDPModule)  # orig module should be composed with FSDPModule
            mod_fsdp_state = m._get_fsdp_state()
            mixed_prec_state = mod_fsdp_state._mp_policy
            if self.pl_module.precision_key:
                assert mixed_prec_state.param_dtype == precision
                assert mixed_prec_state.reduce_dtype == precision
                assert mixed_prec_state.output_dtype == precision
                # assert mixed_prec_state.cast_forward_inputs == True  # not currently inspected
            for fsdp_p in mod_fsdp_state._fsdp_param_group.fsdp_params:
                # test currently assumes 1D sharding on dim 0
                dp_dim0 = self.pl_module.trainer.strategy.device_mesh['data_parallel'].shape[0]
                assert fsdp_p.sharded_size[0] == fsdp_p._orig_size[0] // dp_dim0

    def _collect_fsdp_mod_states(self, fsdp_keys: KeysView) -> Dict[Any, Dict]:
        fsdp_mod_states = {}
        if len(fsdp_keys) > 0:
            for n, m in self.pl_module.named_modules():
                if n in fsdp_keys:
                    fsdp_mod_states.setdefault(n, {})
                    fsdp_mod_states[n]['is_fsdp_managed'] = getattr(m, '_is_fsdp_managed_module', False)
                    fsdp_mod_states[n]['is_fsdp_composed'] = issubclass(type(m), FSDPModule)
                    if fsdp_mod_states[n]['is_fsdp_composed']:
                        mod_fsdp_state = m._get_fsdp_state()
                        fsdp_mod_states[n]['prec_policy_summ'] = (
                            mod_fsdp_state._mp_policy.param_dtype, mod_fsdp_state._mp_policy.reduce_dtype,
                            mod_fsdp_state._mp_policy.output_dtype, mod_fsdp_state._mp_policy.cast_forward_inputs)
                        fsdp_mod_states[n]['param_group_summ'] = (
                            [(fsdp_p._param_fqn, fsdp_p._orig_size, fsdp_p.sharded_size) for fsdp_p in \
                             mod_fsdp_state._fsdp_param_group.fsdp_params])
        return fsdp_mod_states


################################################################################
# Model Parallel Test Models
################################################################################

class FTSBaseModelParallel(FinetuningSchedulerBoringModel):
    def __init__(self, cm_fsdp_plan: Dict, cm_tp_plan: Union[Dict, Callable],
                 module_cls: nn.Module = FTSToyTransformer, loss_parallel: bool = True,
                 tt_cfg: Optional[TestModelArgs] = None,
                 precision_key: Optional[str] = None,
                 monitor_sync_dist: bool = True,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cm_fsdp_plan = cm_fsdp_plan or {}
        self.cm_tp_plan = cm_tp_plan or {}
        self.precision_key = precision_key
        self.tt_cfg = tt_cfg
        self.loss_parallel = loss_parallel
        self.monitor_sync_dist = monitor_sync_dist
        self.module_cls = module_cls
        self.model = self.module_cls(self.tt_cfg)

    def loss_fn(self, output: Tensor, target: Tensor) -> Tensor:
        if self.loss_parallel:
            with loss_parallel():
                loss = F.cross_entropy(output.reshape(-1, output.size(-1)), target.reshape(-1))
        else:
            loss = F.cross_entropy(output.reshape(-1, output.size(-1)), target.reshape(-1))
        return loss

    def forward(self, inputs: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        return self.model(inputs, mask=mask)

    def backward(self, *args, **kwargs):
        if self.loss_parallel:
            with loss_parallel():
                super().backward(*args, **kwargs)
        else:
            super().backward(*args, **kwargs)

    def training_step(self, batch: Tensor, batch_idx: int) -> STEP_OUTPUT:
        inputs, target = batch
        output = self(inputs)
        loss = self.loss_fn(output, target)
        self.training_step_outputs.append(loss)
        return {"loss": loss}

    def validation_step(self, batch: Tensor, batch_idx: int) -> Optional[STEP_OUTPUT]:
        inputs, target = batch
        output = self(inputs)
        # TODO: for now, not using diverge_on_epoch for simplicity
        loss = self.loss_fn(output, target)
        self.validation_step_outputs.append(loss)
        self.log(self.monitor_metric, loss.item(), prog_bar=False, sync_dist=self.monitor_sync_dist)
        return {"x": loss}

    def prepare_data(self) -> None:
        FTSWikiText2(download=True)

    def setup(self, stage):
        self.dataset = FTSWikiText2()

    def get_dataloader(self, *args, **kwargs) -> DataLoader:
        return DataLoader(self.dataset, *args, **kwargs)

    train_dataloader = partialmethod(get_dataloader, batch_size=2)
    val_dataloader = partialmethod(get_dataloader, batch_size=2)
    test_dataloader = partialmethod(get_dataloader, batch_size=2)
    predict_dataloader = partialmethod(get_dataloader, batch_size=2)

class FTSCmModelParallel(FTSBaseModelParallel):
    def __init__(self, leave_auto_composable: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._leave_composable = leave_auto_composable

    def configure_model(self) -> None:
        if self.cm_tp_plan:
            tp_mesh = self.device_mesh["tensor_parallel"]
            if callable(self.cm_tp_plan):
                self.model = self.cm_tp_plan(self.model, tp_mesh, self.loss_parallel)
            else:
                self.model = parallelize_module(self.model, tp_mesh, self.cm_tp_plan)

        if self.cm_fsdp_plan:
            fts_cb = [cb for cb in self.trainer.callbacks if isinstance(cb, (FinetuningScheduler))][0]
            dp_mesh = self.device_mesh["data_parallel"]
            assert dp_mesh.ndim == 1  # Hybrid-sharding not supported

            cm_plan_keys = self.cm_fsdp_plan.keys()
            named_modules = dict(self.named_modules())
            for n in cm_plan_keys:
                if n in named_modules:
                    # we use our auto FSDP plan helpers to configure our manual configure_model plan
                    self.cm_fsdp_plan[n] = fts_cb.strategy_adapter._resolve_cfg_aliases(self.cm_fsdp_plan[n])
                    if act_ckpt := self.cm_fsdp_plan[n].pop('act_ckpt', None):
                        if act_ckpt.mode == ActCkptEnum.COMPOSABLE:
                            act_ckpt._func(self.get_submodule(n), **act_ckpt.cfg)
                        elif act_ckpt.mode in (ActCkptEnum.WRAPPED, ActCkptEnum.WRAPPED_OFFLOAD):
                            module_prefix, _, module_name = n.rpartition('.')
                            outer_mod = self.model if not module_prefix else self.get_submodule(module_prefix)
                            setattr(outer_mod, module_name, act_ckpt._func(self.get_submodule(n), **act_ckpt.cfg))
                    fully_shard(self.get_submodule(n), mesh=dp_mesh, **self.cm_fsdp_plan[n])

################################################################################
# Model Parallel Specially Instrumented FTS Versions
################################################################################

class ModelParallelTestFTS(FSDPStateInspectMixin, TestFinetuningScheduler):

    def __init__(self, ext_tensor_details: Optional[bool] = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ext_tensor_details = ext_tensor_details

    def state_dict(self) -> Dict[str, Any]:
        return super(TestFinetuningScheduler, self).state_dict()

    def restore_best_ckpt(self) -> None:
        super(TestFinetuningScheduler, self).restore_best_ckpt()
        self.restored_best_cnt += 1

    def on_train_epoch_start(self, trainer, pl_module):
        super(TestFinetuningScheduler, self).on_train_epoch_start(trainer, pl_module)
        model_parallel_sample = {}
        state_key = trainer.current_epoch
        if expected_epoch_state := self.expected_state.get(state_key):
            if target_p_keys := expected_epoch_state[0].get('p_states', {}).keys():
                model_parallel_sample['p_states'] = self._collect_p_states(target_p_keys)
            if target_mod_keys := expected_epoch_state[0].get('fsdp_mod_states', {}).keys():
                model_parallel_sample['fsdp_mod_states'] = self._collect_fsdp_mod_states(target_mod_keys)
            if target_mod_keys or target_p_keys:
                current_state = (
                    model_parallel_sample,
                    len(self._fts_state._curr_thawed_params),
                )
                lrs_state = None
                self.inspect_or_assert(current_state, lrs_state, state_key)

    def _collect_p_states(self, tp_keys: KeysView) -> Dict[Any, Dict]:
        p_states = {}
        if len(tp_keys) > 0:
            for n, p in self.pl_module.named_parameters():
                if n in tp_keys:
                    p_states.setdefault(n, {})
                    p_states[n]['requires_grad'] = p.requires_grad
                    p_states[n]['is_DTensor'] = isinstance(p, DTensor)
                    if self.ext_tensor_details:
                        p_states = self._extended_tensor_details(p_states, n, p)
        return p_states

    def _extended_tensor_details(self, p_states: Dict, n: str, p: Tensor) -> None:
        p_states[n]['dtype'] = p.dtype
        p_states[n]['orig_shape'] = p.shape
        if p_states[n]['is_DTensor']:
            p_states[n]['local_shape'] = getattr(p._local_tensor, 'shape', None)
            placement_summ = []
            for pl in p.placements:
                if pl.is_replicate():
                    placement_summ.append('replica')
                elif pl.is_shard():
                    placement_summ.append('shard(dim=' + str(pl.dim) + ')')
                else:
                    placement_summ.append(None)
            p_states[n]['device_mesh'] = DeviceMeshSummary(p.ndim, p.device_mesh.ndim, p.device_mesh.shape,
                                                        p.device_mesh.mesh_dim_names, placement_summ)
        return p_states


################################################################################
# Model Parallel FTS Test Fixtures
################################################################################

@pytest.fixture(scope="module")
def model_parallel_ft_schedule(tmpdir_factory) -> Tuple[Path, Dict]:
    """Generates a default fine-tuning schedule for 'implicit' testing, a modified one for 'explicit' mode and an
    epoch-driven transitions only one for epoch_transitions_only testing."""
    seed_everything(42)
    callbacks = [FinetuningScheduler(gen_ft_sched_only=True)]
    model = FinetuningSchedulerBoringModel()
    tmpdir, rank = get_sched_fixture_tmpdir(tmpdir_factory)
    trainer = Trainer(default_root_dir=tmpdir, callbacks=callbacks, devices=1)
    unmod_schedule_file = tmpdir / "lightning_logs" / "version_0" / f"{model.__class__.__name__}_ft_schedule.yaml"
    # N.B. Though we run this fixture for each rank to avoid adding special logic to each distributed client test, we
    # only generate a schedule on rank 0, linking to it on the other ranks.
    if rank == 0:
        with pytest.raises(SystemExit):
            trainer.fit(model)
    mp_tp_sched_dict = get_fts(trainer).load_yaml_schedule(unmod_schedule_file)
    mp_tp_sched_dict[0]["params"] = [r"model.output.weight", r"model.norm.*"]
    mp_tp_sched_dict[0]["max_transition_epoch"] = 1
    mp_tp_sched_dict[1]["params"] = [r"model.layers.1.(feed_forward|ffn_norm|attention.w.*|attention_norm).*"]
    mp_tp_sched_dict[1]["max_transition_epoch"] = 2
    mp_tp_sched_dict[2]["params"] = [r"model.layers.0.(feed_forward|ffn_norm|attention.w.*|attention_norm).*",
                                     r"model.(pos_embeddings|tok_embeddings).weight"]
    mp_tp_sched_dict.pop(3)
    # TODO: remove this dbg schedule before initial release
    mp_tp_dbg_req_grad = deepcopy(mp_tp_sched_dict)
    mp_tp_dbg_req_grad[0]["params"] = [r"model.layers.1.attention_norm.*", r"model.layers.1.feed_forward.w2.*",
    ]
    mp_tp_dbg_req_grad[1] = {"params": [
        r"model.output.weight",
        r"model.norm.*",
        r"model.layers.[0-1].(ffn_norm|attention.w.*).*",
        r"model.layers.0.attention_norm.*",
        r"model.layers.[0-1].feed_forward.w1.*",
        r"model.layers.0.feed_forward.w2.*",
    ]}
    mp_tp_dbg_req_grad[2] = {"params": [r"model.pos_embeddings.weight",]}

    return (
        unmod_schedule_file,
        mp_tp_sched_dict,
        mp_tp_dbg_req_grad,
    )

# modified version of https://bit.ly/torchtitan_transformer_tp_plan
def gen_apply_transformer_tp_plan(model: nn.Module, device_mesh: DeviceMesh, loss_parallel: bool) -> nn.Module:
    """Apply tensor parallelism."""

    # we're only applying tensor parallelism, composable fsdp is applied subsequently elsewhere if requested
    tp_mesh = device_mesh["tensor_parallel"]

    # 1. Parallelize the embedding and shard its outputs
    # 2. Parallelize the root norm layer over the sequence dim
    non_transformerblock_tp_plan = {
            "tok_embeddings": RowwiseParallel(input_layouts=Replicate(), output_layouts=Shard(1),),
            "pos_embeddings": RowwiseParallel(input_layouts=Replicate(), output_layouts=Shard(0)),
            "norm": SequenceParallel(),
        }

    model = parallelize_module(model, tp_mesh, non_transformerblock_tp_plan)

    # N.B. We're not supporting Float8 parallelism in this test initially (e.g. Float8RowwiseParallel)

    # Apply tensor + sequence parallelism to every transformer block
    for transformer_block in model.layers:
        layer_plan = {
            "attention": PrepareModuleInput(
                input_layouts=Shard(1),
                desired_input_layouts=Replicate(),
            ),
            "attention_norm": SequenceParallel(),
            "attention.wq": ColwiseParallel(use_local_output=False),
            "attention.wk": ColwiseParallel(use_local_output=False),
            "attention.wv": ColwiseParallel(use_local_output=False),
            "attention.wo": RowwiseParallel(output_layouts=Shard(1)),
            "ffn_norm": SequenceParallel(),
            "feed_forward.w1": ColwiseParallel(input_layouts=Shard(1)),
            "feed_forward.w2": RowwiseParallel(output_layouts=Shard(1)),
        }

        parallelize_module(
            module=transformer_block,
            device_mesh=tp_mesh,
            parallelize_plan=layer_plan,
        )

    output_parallelize_plan = (
        ColwiseParallel(
            input_layouts=Shard(1),
            output_layouts=Shard(-1) if loss_parallel else Replicate(),
            use_local_output=not loss_parallel,
        )
    )
    parallelize_module(model.output, tp_mesh, output_parallelize_plan)

    # Manually set output.weight so that parameters and gradients are shared.
    if model.init_args.weight_tying:
        model.output.weight = model.tok_embeddings.weight

    return model


################################################################################
# Model Parallel Test Configuration Aliases
################################################################################

## Model Aliases
fsdp_auto_mod_parallel = FTSBaseModelParallel  # model w/ automatic `configure_model` method (for FSDP)
cm_mod_parallel = FTSCmModelParallel  # model w/ manual `configure_model` method

## toy transformer cfgs
basic_tt = TestModelArgs()
if _TORCH_GREATER_EQUAL_2_5:
    sdp_math_impl_tt = TestModelArgs(avail_sdp_backends=[SDPBackend.MATH], sdp_use_implicit_replication=True)
else:
    sdp_math_impl_tt = None

## DTensor Placement Plan Aliases
cm_tt_tp_plan = gen_apply_transformer_tp_plan

## Auto and Manual FSDP Plan Aliases
fsdp_basic = {'model.output': {}, 'model.layers.1': {}, 'model.norm': {}}
fsdp_cm_shard_tt_basic_act = {'model.output': {'act_ckpt': ('composable', {})},
                              'model.layers.1': {'act_ckpt': ('composable',)}, 'model.norm': {}}
fsdp_auto_tt_basic_act = {'model.output': {'act_ckpt': ('wrapped_offload', {})},
                          'model.layers.1': {'act_ckpt': ('composableX',)},
                          'model.norm': {'act_ckpt': ('wrapped', {})}}
fsdp_autocm_shard_tt = {'^.*model.layers.0$': {}}
fsdp_tp_autocm_shard_tt = {**fsdp_autocm_shard_tt, 'model': {}}
# should warn the second spec cannot be applied and the third spec is a disallowed type
fsdp_autocm_shard_tt_warn = {**fsdp_autocm_shard_tt, 'model.layers.1.ffn_norm': {}, 'model.layers': {},
                            '^.*model.laye.*s$': {}}
# generate an unmatched auto plan spec misconfiguration exception
fsdp_autocm_shard_tt_err = {**fsdp_autocm_shard_tt,'^.*model.nomatch.*s$': {}}

## Model Parallel Model Configuration Aliases
fsdp_cm_only = {"cm_fsdp_plan": fsdp_basic, "cm_tp_plan": None, "module_cls": FTSToyTransformer, "tt_cfg": basic_tt}
fsdp_cm_act = {**fsdp_cm_only, "cm_fsdp_plan": fsdp_cm_shard_tt_basic_act}
tp_only = {"cm_fsdp_plan": None, "cm_tp_plan": cm_tt_tp_plan, "module_cls": FTSToyTransformer, "tt_cfg": basic_tt}
tp_no_lp = {**tp_only, "loss_parallel": False}
tp_lp_math_sdp_impl = {**tp_only, "tt_cfg": sdp_math_impl_tt}
fsdp_tp = {**tp_only, "cm_fsdp_plan": fsdp_basic}
fsdp_auto = {"cm_fsdp_plan": None, "cm_tp_plan": None, "module_cls": FTSToyTransformer, "tt_cfg": basic_tt}

## Model Parallel Strategy Adpater Configuration Aliases
fsdp_auto_only = {"fsdp_plan": fsdp_basic, "fsdp_default_kwargs":{}}
fsdp_auto_cpuoffld = {"fsdp_plan": fsdp_basic, "fsdp_default_kwargs":{'cpu_offload_policy': {}}}
fsdp_auto_act = {"fsdp_plan": fsdp_auto_tt_basic_act, "fsdp_default_kwargs":{}}
fsdp_autocm = {"fsdp_plan": fsdp_autocm_shard_tt_warn, "fsdp_default_kwargs":{}}
fsdp_autocm_err = {"fsdp_plan": fsdp_autocm_shard_tt_err, "fsdp_default_kwargs":{}}
fsdp_autocm_tp = {"fsdp_plan": fsdp_tp_autocm_shard_tt, "fsdp_default_kwargs":{}}

## Model Parallel Strategy Aliases
dp1_tp2 = {"data_parallel_size": 1, "tensor_parallel_size": 2}
dp2_tp1 = {"data_parallel_size": 2, "tensor_parallel_size": 1}

## Lightning Trainer Configuration Aliases
trainer_defaults = {"accelerator": "gpu", "devices": 2, 'limit_train_batches': 2, 'limit_val_batches': 2,
                    'num_sanity_val_steps': 0}

## Precision Configuration Aliases
fp16 = {"precision": "16-true"}
bf16 = {"precision": "bf16-true"}

## Model Parallel Test Configuration Dataclass

# TODO: update this to use dataclass inheritance once python 3.10 is the minimum supported version of python
@dataclass
class ModParallelTestCfg:
    model_cfg_key: str
    model_cls: Callable
    model_cfg: Dict = field(default_factory=dict)
    trainer_cfg: Dict = field(default_factory=lambda: {'max_epochs': 3})
    strategy_cfg: Dict = field(default_factory=lambda: dp2_tp1)
    strategy_adapter_cfg: Dict = field(default_factory=dict)
    precision_opts: Dict = field(default_factory=lambda: {'precision': '32-true'})
    use_dynamo: bool = False
    use_implicit_replication: bool = False
    fts_cls: Callable = ModelParallelTestFTS
    fts_cfg: Dict = field(default_factory=dict)
    ft_sched_idx: int = 1
    es_cls: Callable = FTSEarlyStopping
    es_cfg: Dict = field(default_factory=lambda: {"patience": 1})
    ckpt_cfg: Dict = field(default_factory=lambda: {"save_top_k": 3})
    ckpt_cls: Callable = FTSCheckpoint
    expected_results: ExpectedResults = ExpectedResults()
    runif_alias: Optional[str] = None

    def __post_init__(self):
        self.default_fts_cfg = {
            "logging_level": DEBUG,
            "strategy_adapter_cfg": self.strategy_adapter_cfg,
        }
        default_dep_cfg = {"monitor": "val_loss"}
        self.es_cfg = {**self.es_cfg, **default_dep_cfg}
        self.ckpt_cfg = {**self.ckpt_cfg, **default_dep_cfg}


def test_model_parallel_enapsulated_imports():
    import sys
    from importlib import reload
    modules_to_reload = ('finetuning_scheduler.strategy_adapters._mp_imports',)
    for module_fqn in modules_to_reload:
        del sys.modules[module_fqn]
    with mock.patch("lightning_utilities.core.imports.compare_version", return_value=False):
        from finetuning_scheduler.strategy_adapters._mp_imports import parallelize_module
        assert parallelize_module.__module__ == 'builtins'
    if _TORCH_GREATER_EQUAL_2_5:
        for module_fqn in modules_to_reload:
            reload(sys.modules[module_fqn])
        from finetuning_scheduler.strategy_adapters._mp_imports import parallelize_module
        assert parallelize_module.__module__ == 'torch.distributed.tensor.parallel.api'


@mock.patch("finetuning_scheduler.strategy_adapters.model_parallel._TORCH_GREATER_EQUAL_2_5", False)
def test_torch_greater_equal_2_5():
    with pytest.raises(MisconfigurationException, match="requires PyTorch 2.5 or higher"):
        ModelParallelStrategyAdapter()


## Model Parallel Test Definitions
FTS_MODEL_PARALLEL_PATH_TESTS = (
    # FSDP2 tests
    ModParallelTestCfg(model_cfg_key="fsdp_cm_only", model_cls=cm_mod_parallel, model_cfg=fsdp_cm_only,
                       # use_dynamo=True,  # TODO: test with dynamo once Lightning supports it with ModelParallel
                       runif_alias="alone", expected_results=ExpectedResults(expected_state=path_fsdp)),
    ModParallelTestCfg(model_cfg_key="fsdp_cm_act", model_cls=cm_mod_parallel, model_cfg=fsdp_cm_act,
                       expected_results=ExpectedResults(expected_state=path_fsdp), runif_alias="alone"),
    ModParallelTestCfg(model_cfg_key="fsdp_cm_fp16", model_cls=cm_mod_parallel, precision_opts=fp16,
                       model_cfg=fsdp_cm_only, runif_alias="alone"),
    # FSDP2 auto plan tests
    ModParallelTestCfg(model_cfg_key="fsdp_auto", model_cls=fsdp_auto_mod_parallel, model_cfg=fsdp_auto,
                       strategy_adapter_cfg=fsdp_auto_cpuoffld, runif_alias="alone",
                       expected_results=ExpectedResults(expected_state=path_fsdp)),
    # TODO: upstream bug currently preventing use of `fsdp_auto_cpuoffld` here w/ fp16
    ModParallelTestCfg(model_cfg_key="fsdp_auto_fp16", model_cls=fsdp_auto_mod_parallel, model_cfg=fsdp_auto,
                       precision_opts=fp16, strategy_adapter_cfg=fsdp_auto_only, runif_alias="alone",
                       expected_results=ExpectedResults(expected_state=path_fsdp)),
    ModParallelTestCfg(model_cfg_key="fsdp_auto_act", model_cls=fsdp_auto_mod_parallel,
                       strategy_adapter_cfg=fsdp_auto_act, model_cfg=fsdp_auto, runif_alias="alone",
                       expected_results=ExpectedResults(
                           expected_state=path_fsdp, warns_expected=("Unknown AC mode", "to filter out AC"))),
    ModParallelTestCfg(model_cfg_key="fsdp_autocm", model_cls=cm_mod_parallel, strategy_adapter_cfg=fsdp_autocm,
                       model_cfg=fsdp_cm_only, use_implicit_replication=True, runif_alias="alone",
                       expected_results=ExpectedResults(
                           expected_state=path_fsdp_autocm,
                           warns_expected=("parents has already registered", "has the unsupported type"))),
    ModParallelTestCfg(model_cfg_key="fsdp_autocm_unmatched", model_cls=cm_mod_parallel, model_cfg=fsdp_cm_only,
                       strategy_adapter_cfg=fsdp_autocm_err, runif_alias="alone", use_implicit_replication=True,
                       expected_results=ExpectedResults(expected_state=path_fsdp_autocm,
                                                        exceptions_expected=("did not match any named modules",))),
    # TP tests
    ModParallelTestCfg(model_cfg_key="tp_only", model_cls=cm_mod_parallel, model_cfg=tp_no_lp, strategy_cfg=dp1_tp2,
                       runif_alias="alone", expected_results=ExpectedResults(expected_state=path_tp)),
    ModParallelTestCfg(model_cfg_key="tp_lp_fp16", model_cls=cm_mod_parallel, precision_opts=fp16, model_cfg=tp_only,
                       strategy_cfg=dp1_tp2, runif_alias="alone",
                       expected_results=ExpectedResults(expected_state=path_tp)),
    # TODO: given that SDPBackend.MATH with lp causes (at least without special accommodation)
    #       https://github.com/pytorch/pytorch/pull/149764 as of PT 2.8, we are disabling this test until our
    #       infrastructure is upgraded and we can use SDPBackend.FLASH with bf16 on all our cards
    # ModParallelTestCfg(model_cfg_key="tp_lp_bf16", model_cls=cm_mod_parallel, precision_opts=bf16,
    #                    model_cfg=tp_lp_math_sdp_impl, strategy_cfg=dp1_tp2, runif_alias="bf16_alone"),
    # FSDP2 + TP (trivial submesh) tests
    # temporarily disabling this test as it triggers a hang in `fsdp_autocm_tp` about 10% of the time and
    # `fsdp_autocm_tp` and the marginal utility of `fsdp_tp` is minimal
    # ModParallelTestCfg(model_cfg_key="fsdp_tp", model_cls=cm_mod_parallel, model_cfg=fsdp_tp,
    # runif_alias="einsum_exp", expected_results=ExpectedResults(expected_state=path_tp_fsdp)),
    ModParallelTestCfg(model_cfg_key="fsdp_autocm_tp", model_cls=cm_mod_parallel, strategy_adapter_cfg=fsdp_autocm_tp,
                       model_cfg=fsdp_tp, runif_alias="einsum_exp",
                       expected_results=ExpectedResults(expected_state=path_tp_fsdp_autocm)),
)
@RunIf(min_cuda_gpus=2, min_torch="2.5.0")
@pytest.mark.parametrize("test_cfg", pytest_param_factory(FTS_MODEL_PARALLEL_PATH_TESTS))
def test_fts_model_parallel_integration(tmpdir, recwarn, model_parallel_ft_schedule, test_cfg):
    """Validate :class:`~finetuning_scheduler.FinetuningScheduler` functions properly in a supported 'ddp'
    distributed context."""
    seed_everything(42)
    # one can manually set this to True for a local test override
    state_log_dir = tmpdir if FTS_GLOBAL_STATE_LOG_MODE else None
    ft_sched = model_parallel_ft_schedule[test_cfg.ft_sched_idx]
    callbacks = callbacks_cfg(ft_sched, state_log_dir, test_cfg)
    strategy = ModelParallelStrategy(**test_cfg.strategy_cfg)
    trainer_config = {"strategy": strategy, "callbacks": callbacks, "default_root_dir": tmpdir, **trainer_defaults,
                      **test_cfg.trainer_cfg, **test_cfg.precision_opts}
    trainer = Trainer(**trainer_config)
    with trainer.init_module(empty_init=True):
        model = test_cfg.model_cls(**test_cfg.model_cfg)
    configured_model = torch.compile(model) if test_cfg.use_dynamo else model
    if exc_expect := test_cfg.expected_results.exceptions_expected:
        gen_exceptions(trainer, configured_model, exc_expect)
    else:
        if test_cfg.use_implicit_replication:
            with implicit_replication():
                trainer.fit(configured_model)
        else:
            trainer.fit(configured_model)
        default_fts_sanity_chk(trainer)
    if trainer.is_global_zero:
        fts_check_warns(recwarn, expected_warns=MODEL_PARALLEL_BASE_WARNS,
                        warns_expected=test_cfg.expected_results.warns_expected,
                        expected_warns_dynamo=MODEL_PARALLEL_DYNAMO_EXPECTED_WARNS, use_dynamo=test_cfg.use_dynamo)

def gen_exceptions(trainer, model, exception_expected):
    if isinstance(exception_expected, tuple):
        exception_expected = "|".join(exception_expected)
    with pytest.raises(MisconfigurationException, match=exception_expected):
            trainer.fit(model)


def callbacks_cfg(ft_sched, state_log_dir, test_cfg):
    active_fts_cfg = {"ft_schedule": ft_sched, "strategy_adapter_cfg": test_cfg.strategy_adapter_cfg,
                      "expected_state": test_cfg.expected_results.expected_state, "state_log_dir": state_log_dir,
                      **test_cfg.fts_cfg, }
    callbacks = [test_cfg.fts_cls(**active_fts_cfg)]
    for tcls, tcfg, subc in zip((test_cfg.es_cls, test_cfg.ckpt_cls), (test_cfg.es_cfg, test_cfg.ckpt_cfg),
                                (FTSEarlyStopping, FTSCheckpoint)):
        if issubclass(tcls, subc):
            callbacks.append(tcls(**tcfg))
    return callbacks



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Fine-Tuning Scheduler Model Parallel Strategy Adapter
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

A :class:`~finetuning_scheduler.strategy_adapters.StrategyAdapter` that extends Fine-Tuning Scheduler's support
for PyTorch's distributed composable and Tensor Parallel APIs.

"""
from typing import Any, Optional, Callable, Dict, Sequence, Tuple, List, Union
from functools import wraps
from copy import deepcopy
import re
import os
from pprint import pformat
from dataclasses import dataclass, field

import torch
from lightning.fabric.utilities.enums import LightningEnum
from lightning.pytorch.utilities.exceptions import MisconfigurationException
from lightning.pytorch.utilities.model_helpers import is_overridden
from lightning.fabric.utilities import rank_zero_warn, rank_zero_info
from lightning.pytorch.utilities.rank_zero import rank_zero_debug

from finetuning_scheduler.strategy_adapters.base import StrategyAdapter
from finetuning_scheduler.strategy_adapters._wrap_utils import _compose_ncac
# conditionally import indirectly to avoid duplicating import logic in several different modules
from finetuning_scheduler.strategy_adapters._mp_imports import (_TORCH_GREATER_EQUAL_2_5, DTensor, FSDPModule,
                                                                fully_shard, CPUOffloadPolicy, checkpoint,
                                                                checkpoint_wrapper, ActivationWrapper, offload_wrapper)

class ActCkptEnum(LightningEnum):
    COMPOSABLE = "composable"
    WRAPPED = "wrapped"
    WRAPPED_OFFLOAD = "wrapped_offload"

@dataclass
class ActCkptCfg:
    mode: Union[ActCkptEnum, str]
    cfg: Optional[Dict[str, Any]] = field(default_factory=dict)
    _func: Optional[Callable] = None

    def __post_init__(self) -> None:
        if isinstance(self.mode, str):
            try:
                self.mode = ActCkptEnum(self.mode)
            except ValueError:
                rank_zero_warn(f"Unknown AC mode: {self.mode}. Defaulting to `{ActCkptEnum.COMPOSABLE.value}` mode.")
                self.mode = ActCkptEnum.COMPOSABLE
        if self.mode == ActCkptEnum.COMPOSABLE:
            self._func = checkpoint
        elif self.mode == ActCkptEnum.WRAPPED_OFFLOAD:
            self._func = offload_wrapper
        elif self.mode == ActCkptEnum.WRAPPED:
            self._func = checkpoint_wrapper


class ModelParallelStrategyAdapter(StrategyAdapter):
    r"""
    A :class:`~finetuning_scheduler.strategy_adapters.StrategyAdapter` that extends
    :class:`~finetuning_scheduler.fts.FinetuningScheduler` (FTS) to support flexible, multi-phase, scheduled fine-tuning
    with PyTorch's composable distributed (e.g. ``fully_shard``) and Tensor Parallelism APIs.
    FTS augments Lightning's Model Parallel strategy
    (:external+pl:class:`~lightning.pytorch.strategies.model_parallel.ModelParallelStrategy`) by allowing users to apply
    the ``fully_shard`` API using module name/pattern-based configuration instead of manually inspecting modules and
    applying the API in ``LightningModule.configure_model`` (see
    :attr:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter.fsdp_plan`).

    See the :ref:`model-parallel-fine-tuning-examples` tutorial for a concrete example and additional guidance.

    .. warning::
        :class:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter` is in BETA and subject to change.
        The interface can bring breaking changes and new features with the next release of PyTorch.

    .. note::
        ``fsdp_plan`` module name/pattern-based ``fully_shard`` directives are applied after any preceding Tensor
        Parallel or explicit ``fully_shard`` directives in ``LightningModule.configure_model``. FTS will only apply
        ``fully_shard`` to a specified module if it was not already applied to that module.

    .. note::
        In addition to all valid ``fully_shard`` API kwargs, ``fsdp_plan`` also supports a ``act_ckpt`` and
        ``cpu_offload_policy`` kwargs.

    For specified module/patterns (or ``fsdp_default_kwargs``), ``act_ckpt`` allows one
    to pass a string alias specifying the use of the desired activation checkpointing API (e.g. "composable",
    "wrapped", "wrapped_offload") as well as an optional ``Dict`` of activation checkpointing kwargs. The specified
    checkpointing APIs will be applied to the matching module(s) before ``fully_shard``.

    ``cpu_offload_policy`` is a convenience alias that will apply CPUOffloadPolicy to the matching module(s) along
    with any provided ``Dict`` of policy kwargs.
    """

    def __init__(self, fsdp_default_kwargs: Optional[Dict] = None, fsdp_plan: Optional[Dict] = None, *args: Any,
                 **kwargs: Any) -> None:
        """The only user-facing configuration for
        :class:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter` are
        :attr:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter.fsdp_plan` and
        :attr:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter.fsdp_default_kwargs`.

        Args:
            fsdp_plan (Optional[Dict]): An optional dictionary of module names or regex pattern keys with associated
                ``fully_shard`` composable distributed API kwargs to apply to matching modules.

                - Allows users to apply the ``fully_shard`` API using module name/pattern-based configuration instead of
                  manually inspecting modules and applying the API in ``LightningModule.configure_model``.
                - ``fsdp_plan`` directives can also be composed with explicit ``fully_shard`` calls in
                  ``LightningModule.configure_model``, as the ``fsdp_plan`` directives will only invoke ``fully_shard``
                  on a specified module if it was not already applied to that module.
                - All valid ``fully_shard`` API kwargs are supported.
                - :attr:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter.fsdp_plan` directives are
                  applied in the order provided in the ``fsdp_plan`` dictionary.

                Additionally, ``fsdp_plan`` supports ``act_ckpt`` and ``cpu_offload_policy`` kwargs. For specified
                module/patterns (or ``fsdp_default_kwargs``):

                - ``act_ckpt`` (``Sequence`` [ ``str``, ``Dict`` | ``None`` ] | ``ActCkptCfg``): pass an alias
                  specifying the use of the desired activation checkpointing API (e.g. "composable", "wrapped",
                  "wrapped_offload") as well as an optional ``Dict`` of activation checkpointing kwargs. The specified
                  checkpointing APIs will be applied to the matching module(s) before ``fully_shard``.
                - ``cpu_offload_policy`` (``Dict`` [ ``Optional`` [ ``str`` , ``Any`` ]]) is a convience alias that will
                  apply ``CPUOffloadPolicy`` to the matching module(s) along with any provided Dict of policy kwargs.
                  Defaults to ``None``.

            fsdp_default_kwargs (Optional[Dict]): An optional dictionary of default ``fully_shard`` API kwargs to apply
                to each matching module in ``fsdp_plan``. Module-name/pattern specific kwargs will take precedence over
                these. All kwargs valid for ``fsdp_plan`` above are supported. Defaults to ``None``.

        Attributes:
            fsdp_plan: An optional dictionary of module names or regex pattern keys with associated
                ``fully_shard`` composable distributed API kwargs to apply to matching modules.

                - Allows users to apply the ``fully_shard`` API using module name/pattern-based configuration instead of
                  manually inspecting modules and applying the API in ``LightningModule.configure_model``.
                - ``fsdp_plan`` directives can also be composed with explicit ``fully_shard`` calls in
                  ``LightningModule.configure_model``, as the ``fsdp_plan`` directives will only invoke ``fully_shard``
                  on a specified module if it was not already applied to that module.
                - All valid ``fully_shard`` API kwargs are supported.
                - :attr:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter.fsdp_plan` directives are
                  applied in the order provided in the ``fsdp_plan`` dictionary.

                Additionally, ``fsdp_plan`` supports ``act_ckpt`` and ``cpu_offload_policy`` kwargs. For specified
                module/patterns (or ``fsdp_default_kwargs``):

                - ``act_ckpt`` (``Sequence`` [ ``str``, ``Dict`` | ``None`` ] | ``ActCkptCfg``): pass an alias
                  specifying the use of the desired activation checkpointing API (e.g. "composable", "wrapped",
                  "wrapped_offload") as well as an optional ``Dict`` of activation checkpointing kwargs. The specified
                  checkpointing APIs will be applied to the matching module(s) before ``fully_shard``.
                - ``cpu_offload_policy`` (``Dict`` [ ``Optional`` [ ``str`` , ``Any`` ]]) is a convience alias that will
                  apply ``CPUOffloadPolicy`` to the matching module(s) along with any provided Dict of policy kwargs.

            fsdp_default_kwargs: An optional dictionary of default ``fully_shard`` API kwargs to apply to each matching
                module in ``fsdp_plan``. Module-name/pattern specific kwargs will take precedence over
                these. All kwargs valid for ``fsdp_plan`` above are supported.
        """
        super().__init__(*args, **kwargs)
        self.fsdp_default_kwargs = fsdp_default_kwargs or {}
        self.fsdp_plan = fsdp_plan or {}
        if not _TORCH_GREATER_EQUAL_2_5:
            # specifically, depends upon https://github.com/pytorch/pytorch/pull/133502 among other changes
            raise MisconfigurationException(f"{type(self).__name__} requires PyTorch 2.5 or higher.")

    def on_before_init_fts(self) -> None:
        """In this hook executed immediately before
        :meth:`~finetuning_scheduler.fts_supporters.ScheduleImplMixin.init_fts`, to accommodate enhanced Model Parallel
        functionality, we:

        1. Validate the :attr:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter.fsdp_plan`
           configuration
        2. Configure FTS wrapping of the provided :external+pl:class:`~lightning.pytorch.core.module.LightningModule`
           to either use the provided ``LightningModule.configure_model`` method (if present) or a provided
           ``fsdp_plan``.
        """
        self._validate_fsdp_plan()
        if is_overridden("configure_model", self.pl_module):
            rank_zero_debug("Overridden `LightningModule.configure_model` hook identified. Any name-driven distributed "
                            "API plans will be applied after the configure_model hook.")
            cm_func = self._wrapped_configure_model(self.pl_module.configure_model)
            setattr(self.pl_module, "configure_model", cm_func)
        else:
            setattr(self.pl_module, "configure_model", self._fts_auto_configure_model)

    def _validate_fsdp_fts_config(self) -> Optional[Tuple]:
        mixed_param_pgs: dict = {}
        inspection_map = deepcopy(self.fts_handle.ft_schedule)
        all_params = dict(self.pl_module.named_parameters())
        has_mixed_pgs = False
        for d, pl in inspection_map.items():
            mixed_param_pgs[d] = {}
            non_dtensor_params, dtensor_params = [], []
            for lp in pl["params"]:
                param = all_params[lp]
                if isinstance(param, DTensor):
                    dtensor_params.append(lp)
                else:
                    non_dtensor_params.append(lp)
            if non_dtensor_params and dtensor_params:
                has_mixed_pgs = True
                for lp in non_dtensor_params:
                    p = all_params[lp]
                    mixed_param_pgs[d][lp] = type(p.data) if isinstance(p, torch.nn.Parameter) else type(p)

        if has_mixed_pgs:
            ModelParallelStrategyAdapter._provide_mixed_pg_feedback(mixed_param_pgs)

    @staticmethod
    def _provide_mixed_pg_feedback(mixed_param_pgs: Dict) -> None:
        phasewise_feedback = ""
        for d, mixed_params in mixed_param_pgs.items():
            if mixed_params:
                phasewise_feedback += (
                    f"  Phase {d}:  {os.linesep}    {pformat(mixed_params, indent=4)}{os.linesep}"
                )
        if phasewise_feedback:
            summary_msg_base = (
                "\nThe current fine-tuning schedule and FSDP plan produce mixed DTensor/Non-DTensor parameter group(s)"
                " in at least one phase.\n"
                "Be aware some optimizer operations may require you to either:\n"
                "  1. allow implicit replication (experimental)\n"
                "  2. add modules with optimized non-DTensor (usually ``torch.Tensor``) parameters to the FSDP plan.\n"
                "     (HINT: adding your root model to the FSDP plan usually addresses this concern for all phases if"
                " needed).\n\n"
                "Found Non-DTensor parameters in mixed parameter groups: \n"
            )
            summary_msg = summary_msg_base + phasewise_feedback
            rank_zero_info(summary_msg)

    def on_before_fts_fit_start(self) -> None:
        """In this hook executed immediately before the :class:`~finetuning_scheduler.fts.FinetuningScheduler`
        :meth:`~finetuning_scheduler.fts.FinetuningScheduler.on_fit_start` hook begins, we ensure the provided
        fine-tuning schedule and FSDP2 composed :external+pl:class:`~lightning.pytorch.core.module.LightningModule` are
        appropriately aligned. If the fine-tuning schedule and composed modules yield parameter group configurations
        that may not be supported by some optimizer group operations, detailed feedback on potential remediation is
        provided to the user.
        """
        self._validate_fsdp_fts_config()

    def _maybe_update_fsdp_plan(self, resolved_modules: Dict, kwargs: Dict, name: str) -> bool:
        # right now, the only explicitly disallowed types are the container types w/o a forward: `ModuleList``
        #  and `ModuleDict``
        # NB there are other ``fully_shard`` distributed API constraints that the user may need to validate, for
        # instance, shared parameters must belong to the same ``fully_shard`` instance
        # (https://bit.ly/fsdp_dapi_shared_constraint)
        if isinstance(self.pl_module.get_submodule(name), (torch.nn.ModuleList, torch.nn.ModuleDict)):
            rank_zero_warn(f"Provided FSDP module plan includes a directive to apply the FSDP API to ('{name}') which"
                           f" has the unsupported type '{type(self.pl_module.get_submodule(name))}'. This"
                           " directive will be pruned and ignored.")
            return False
        kwargs = ModelParallelStrategyAdapter._resolve_cfg_aliases(kwargs)  # default policy aliases already resolved
        resolved_modules[name] = {**self.fsdp_default_kwargs, **kwargs}
        return True

    @staticmethod
    def _resolve_cfg_aliases(config_dict: Dict) -> Dict:
        """Resolve any provided convenience FSDP default kwargs."""
        for k, v in list(config_dict.items()):
            # currently adding a `cpu_offload_policy` option alias for convenience
            # open a GitHub issue if you think other poliy alias options would be useful
            if k == "cpu_offload_policy":
                config_dict["offload_policy"] = CPUOffloadPolicy(**v)
                del config_dict[k]
            elif k == "act_ckpt":
                if not isinstance(v, ActCkptCfg):
                    config_dict["act_ckpt"] = ActCkptCfg(*v) if isinstance(v, Sequence) else ActCkptCfg(mode=v)
        return config_dict

    def _validate_fsdp_plan(self) -> None:
        """Expand any regex expressions specified in
        :attr:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter.fsdp_plan`.

        Raises:
            MisconfigurationException: If a specified module name or regex does not resolve to at least one named
                module.
        """
        if not self.fsdp_plan:
            return
        if self.fsdp_default_kwargs:
            self.fsdp_default_kwargs = ModelParallelStrategyAdapter._resolve_cfg_aliases(self.fsdp_default_kwargs)
        named_modules = dict(self.pl_module.named_modules()).keys()
        resolved_modules: Dict[str, Dict] = {}
        for plan_n, kwargs in self.fsdp_plan.items():
            module_resolved = False
            if plan_n in named_modules:
                module_resolved = True
                if not self._maybe_update_fsdp_plan(resolved_modules, kwargs, plan_n):
                    continue  # resolution success, but skipping an unsupported module
            else:
                mpat = re.compile(fr"{plan_n}")
                for n in named_modules:
                    if mpat.match(n):
                        module_resolved = True
                        if not self._maybe_update_fsdp_plan(resolved_modules, kwargs, n):
                            continue  # resolution success, but skipping an unsupported module
            if not module_resolved:
                raise MisconfigurationException(
                    f"The module or regex '{plan_n}' specified in `fsdp_plan` did not match any named modules "
                    "in the provided model."
                )
        self.fsdp_plan = resolved_modules

    def _fts_auto_configure_model(self) -> None:
        """A wrapper for configuration-driven composable distributed API configurations that may be composed with
        or substitute for explicit directives in a `LightningModule.configure_model` method.

        Currently, FTS only supports configuration-driven distributed API plans for `fully_shard`. See :attr:`fsdp_plan`
        for details.
        """
        # TODO: replace directly with `_apply_fsdp_plan` if foregoing `tp_auto_plan` implementation
        # self._apply_tp_auto_plan()
        self._apply_fsdp_plan()

    def _compose_or_warn(self, n: str) -> Optional[str]:
        # NB PL does not currently support passing user-created meshes so we unconditionally pass the device mesh
        if not getattr(self.pl_module.get_submodule(n), '_is_fsdp_managed_module', False):
            if act_ckpt := self.fsdp_plan[n].pop('act_ckpt', None):
                self._apply_activation_checkpointing(act_ckpt,n)
            else:
                fully_shard(self.pl_module.get_submodule(n), mesh=self.pls_handle.device_mesh["data_parallel"],
                            **self.fsdp_plan[n])
        else:
            rank_zero_warn(
                f"Module '{n}' or one of its parents has already registered the ``fully_shard`` composable "
                f"distributed API. Applying the ``fully_shard`` API to '{n}' is not supported in this context."
                f"Please apply ``fully_shard`` to '{n}' before any of its parents if it is intended to have a "
                f"distinct `FSDPState`."
            )
            return n  # return the uncomposed fqn for subsequent handling

    def _apply_activation_checkpointing(self, act_ckpt: ActCkptCfg, n: str) -> None:
        assert isinstance(act_ckpt, ActCkptCfg)
        assert isinstance(act_ckpt.cfg, dict) and callable(act_ckpt._func)
        if act_ckpt.mode == ActCkptEnum.COMPOSABLE:
            act_ckpt._func(self.pl_module.get_submodule(n), **act_ckpt.cfg)
        elif act_ckpt.mode in (ActCkptEnum.WRAPPED, ActCkptEnum.WRAPPED_OFFLOAD):
            module_prefix, _, module_name = n.rpartition('.')
            outer_mod = self.pl_module if not module_prefix else self.pl_module.get_submodule(module_prefix)
            setattr(outer_mod, module_name, act_ckpt._func(self.pl_module.get_submodule(n), **act_ckpt.cfg))
            assert isinstance(self.pl_module.get_submodule(n), ActivationWrapper)
        fully_shard(self.pl_module.get_submodule(n), mesh=self.pls_handle.device_mesh["data_parallel"],
                    **self.fsdp_plan[n])

    def _any_noncomposable_AC(self) -> Optional[Tuple[str, type]]:
        for n, m in self.pl_module.named_modules():
            if isinstance(m, ActivationWrapper):
                for c in type(m).__mro__:
                    if issubclass(c, ActivationWrapper) and not issubclass(c, FSDPModule):
                        return n, c
        return None

    def _apply_fsdp_plan(self) -> None:
        """Apply the FSDP distributed API according to the name-driven plan provided by the user via
        :attr:`~finetuning_scheduler.strategy_adapters.FSDPStrategyAdapter.fsdp_plan`. This method can be used in
        conjunction with a user-overridden `LightningModule.configure_model` method or without one. Any name-driven
        FSDP API plan directives will be applied after `configure_model` is executed.

        Note: If the provided `fsdp_plan` specifies a module that has already had ``fully_shard`` applied (e.g.
        explicitly via `configure_model`), that directive will be skipped with a warning.
        """
        auto_plan_keys = self.fsdp_plan.keys()
        named_modules = dict(self.pl_module.named_modules())
        has_fsdp_managed = False
        unsuccessful_compositions: List[str] = []
        for n in auto_plan_keys:
            if n in named_modules:
                unsuccessful_fqn = self._compose_or_warn(n)  # returns fqn if composition could not be performed
                if unsuccessful_fqn:
                    unsuccessful_compositions.append(unsuccessful_fqn)
            if not has_fsdp_managed:
                has_fsdp_managed = getattr(named_modules[n], '_is_fsdp_managed_module', False)
        for k in unsuccessful_compositions:
            del self.fsdp_plan[k]  # user has been warned, remove the unsuccessful directives from the active plan
        if non_comp_ac := self._any_noncomposable_AC():
            _compose_ncac(self.pl_module)
            rank_zero_warn(
                "Non-composable activation checkpoint (NCAC) APIs are being used (e.g. `{}` is `{}`). To better "
                "integrate these APIs with the ``fully_shard`` composable distributed API, FTS has composed the "
                "provided LightningModule with a simple adapter to filter out AC-specific parameter prefixes. This is "
                "an experimental feature and may be removed in a future release.\n"
                "If you don't need specific features of the used non-composable AC APIs, using the composable AC API "
                "is recommended instead, i.e. `act_ckpt=('composable', ...)`.".format(non_comp_ac[0], non_comp_ac[1])
            )

    def _wrapped_configure_model(self, cm_func: Callable) -> Callable:
        """If the user has overridden ``configure_model`` in their ``LightningModule``, wrap the user's
        explicit wrapping method with the required
        :class:`~finetuning_scheduler.strategy_adapters.ModelParallelStrategyAdapter` methods.

        Args:
            cm_func (Callable): The user's overridden ``LightningModule.configure_model`` method

        Returns:
            Callable: The user's overridden ``LightningModule.configure_model`` method wrapped with this
            adapter's internal implementation methods.
        """

        @wraps(cm_func)
        def wrapped_func() -> None:
            cm_func()
            self._post_cm_plan_apply()

        return wrapped_func

    def _post_cm_plan_apply(self) -> None:
        # TODO: determine if we want to provide tp_auto_plan functionality or require manual configure_model for it
        # self.tp_auto_plan()
        if self.fsdp_plan:
            self._apply_fsdp_plan()



"""
FTS Profiling Utilities
=======================

Collection of utilities that expedite simultaneous configuration and orchestration of numerous complementary profiling
methods.

"""
from fts_examples.profiling.config import MemProfilerHooks, MemProfilerSchedule, MemProfilerFuncs, MemProfilerCfg
from fts_examples.profiling.memprofiler import MemProfiler
from fts_examples.profiling.profiler_hooks_mixin import ProfilerHooksMixin
from fts_examples.profiling.extended_profiler import ExtendedPyTorchProfiler

__all__ = [
    'MemProfiler',
    'MemProfilerHooks',
    'MemProfilerSchedule',
    'MemProfilerFuncs',
    'MemProfilerCfg',
    'ExtendedPyTorchProfiler',
    'ProfilerHooksMixin',
]
